# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module contains utilities for manipulating numerical simulation ensembles
# using command line utilities.  This code is modified from the romans toolbox at
# https://github.com/sandialabs/romans.

# See docs/build/html/index.html file for documentation.

# S. Martin
# 1/8/2024


# standard library imports

# parsing command line
import argparse

# logging to file/screen
import logging
import sys

# parsing %d[::] format
import re

# import module from file
import importlib
import pkgutil
import os
from pathlib import Path

# 3rd party imports

# meshio is default file reader/writer
import meshio

# miscellaneous array computations
import numpy as np

# parallel comutation
from ipyparallel import Client

# local imports
import slypi.ensemble.plugins as plugins

# defaults
DEFAULT_LOG_LEVEL = 'info'
DEFAULT_PLUGIN = 'ps'


# set up argument parser with logging
class ArgumentParser(argparse.ArgumentParser):
    """
    Return an instance of argparse.ArgumentParser, pre-configured with arguments
    used to run ensemble tools.
    
    Command line flags defined by this class are:

    * -\\-log-level
    * -\\-log-file
    * -\\-plugin

    The ``--log-level`` flag specifies the log level to use for the screen, the
    ``--log-file`` flag specifies a log file to write (debug and above), and the
    ``--plugin`` flag specifies the plugin Python file to use.

    :Example:

    .. code-block:: python

        import slypi.ensemble

        # describe and initialize parser
        my_description = "My extension of the parameter space parser."
        my_parser = ensemble.ArgumentParser(description=my_description)

        # add an argument
        my_parser.add_argument('--my-flag', help="My extension command line flag")
    """

    # create argument parser with logging level and log file
    def __init__(self, *arguments, **keywords):

        argparse.ArgumentParser.__init__(self, *arguments, **keywords)

        # logging options
        self.add_argument("--log-level", default=DEFAULT_LOG_LEVEL, 
            choices = ["debug", "info", "warning", "error", "critical"], 
            help = "Log level. Default: '%(default)s'")
        self.add_argument('--log-file', type=str,
            help = "Log to file. Notes: (1) If this file already exists it will be "
                 + "overwritten, (2) Log file includes time stamp and is set to "
                 + "debug level")

        # plugin identification
        self.add_argument("--plugin", default=DEFAULT_PLUGIN,
            help = "Plugin Python file name to import (defaults to '%(default)s'), "
                   "can be either a plugin from slypi/ensemble/plugins (no extension) "
                   'or a python file (.py extension). Use "python -m '
                   'slypi.ensemble.plugins.plugin --help" to see any command line '
                   'options for the plugin.')

    # parse logging arguments as well as additional arguments
    def parse_args(self, list_input=None):
        """
        Extends standard argparse.parse_args() call.  Uses parse_known_args()
        to parse the base arguments and any additional arguments.  Returns any
        unknown arguments as a list.  The unknown arguments can be used to
        set plugin specific variables.

        :Example:

        .. code-block:: python

            # parse command line
            args = my_parser.parse_args()

            # parse command line and start logger
            args, unknown_arg_list = my_parser.parse_args()

            # arguments can be accessed using, e.g.
            print(args.log_level)

            # unknown arguments are returned as a list
            print(unknown_arg_list)
            
        :Warning: 
        
        Python uses prefix-matching, so that if a plugin uses an 
        argument flag that matches the prefix of an already existing argument
        (including any argument defined in a utility), that argument will 
        be not be passed onto the plugin.
        """

        # parse arguments
        return argparse.ArgumentParser.parse_known_args(self, list_input)


# starts slypi ensemble logger
def init_logger(log_file=None, log_level=DEFAULT_LOG_LEVEL):
    """
    Starts slypi ensemble logger, sets logging level for console and opens log 
    file, if desired.  Console only outputs messages while log file includes time 
    stamps and origin of message.  Log file is set to debug and above.

    :Example:

    .. code-block:: python

        import logging

        # start logger for extension
        my_log = logging.getLogger("slypi.my_extension")
        my_log.info("My log message.")

        # or use slypi ensemble logger
        slypi.ensemble.log.info("My message.")
    """

    # convert log level to logging class attribute
    log_level_attr = getattr(logging, log_level.upper())

    # set up slypi ensemble logger
    global log
    log = logging.getLogger(__name__)
    
    # log to file, if requested
    if log_file:
        logging.basicConfig(level=logging.DEBUG,
                format='%(asctime)s [%(name)s] %(levelname)s: %(message)s',
                datefmt='%m-%d %H:%M',
                filename=log_file,
                filemode='w',
                force=True)

        # also write messages to console
        console = logging.StreamHandler()
        formatter = logging.Formatter('%(message)s')
        console.setFormatter(formatter)
        console.setLevel(log_level_attr)

        # add console
        log.addHandler(console)

    # otherwise only log to console
    else:
        logging.basicConfig(level=log_level_attr,
                            format='%(message)s')

    log.debug("Started slypi ensemble log.")


# template for plugin architecture to accomodate different 
# i/o formats, ML algorithms, and simulations
class PluginTemplate:
    """
    Provides an extensible architecture for accomodating different input/output
    formats, machine learning algorithms, and simulations.

    Plugins must be defined as following in a seperate ``.py`` file:
    
    :Example:

    .. code-block:: python
    
        import slypi

        class Plugin(slypi.ensemble.PluginTemplate):
            
            ...

    See ``memphis.py`` for an example.
    """

    # initializes plugin by setting up argument parser with
    # any additional command line flags
    def __init__(self, description=None):

        # initialize plugin parser
        self.parser = argparse.ArgumentParser(description=description)

        # add parser arguments
        self.add_args()

        # start plugin logger
        self.log = logging.getLogger("slypi.ensemble.plugin")
        self.log.debug("Started plugin.")

    # add any extra command line arguments
    def add_args(self):
        """
        Note these flags should not conflict with already used flags (see
        parse_args class ArgumentParser).

        :Example:

        .. code-block:: python

            # plugin adds command line argument
            self.parser.add_argument("--my_option", help="My option for plugin.")
        """

        pass

    # parse any plugin specific arguments
    def parse_args(self, arg_list=[]):
        """
        Parses arguments specific to plugin.

        Args:
            arg_list (list): list of command line flags and arguments

        Returns:
            args (object), unknown_args (list): `ArgumentParser` processed argument list,
            list of un-recognized arguments
        """

        # parse only known arguments, ignore others
        args, arg_list = self.parser.parse_known_args(arg_list)

        return args, arg_list

    # check arguments, error if incorrect input
    def check_args(self, args):
        """
        Checks plugin arguments and raises exceptions if there are errors.

        Args:
            args (`ArgumentParser` object): processed argument list
        """

        pass

    # initialize any local variables from command line arguments
    def init(self, args):
        """
        Initialize any local variables from command line arguments.

        Args:
            args (`ArgumentParser` object): processed argument list

        :Example:

        .. code-block:: python

            self.my_var = args.my_option
        """

        pass

    # generic reader
    def read_file(self, file_in, file_type=None):
        """
        Reads a file associated with a single time step in an ensemble.
        File type is inferred from extension unless provided.

        Args:
            file_in (string): name of file to read
            file_type (string): file input type (regardless of extension)

        Returns:
            data (object): file contents
        """

        # read file using meshio
        mesh = meshio.read(file_in, file_type)

        # some converters (like VTK) require `points` to 
        # be contiguous (from meshio-convert)
        mesh.points = np.ascontiguousarray(mesh.points)

        return mesh

    # read files in batches (can run in parallel)
    def read_file_batch(self, batch_files, file_type=None, parallel=False, flatten=True):
        """
        Reads a batch of files from an ensemble.  File type is inferred from
        extension unless provided.  Can be run in parallel using ipyparallel.

        Args:
            batch_files (list): list of files to read
            file_type (string): file input type (regardless of extension)
            parallel (boolean): to run in parallel using ipyparallel
            flatten (boolean): flatten matrix files to vectors (defaults True)

        Returns:
            data_list (list): file of file contents
        """

        # if not parallel, just get all sim files normally
        if not parallel:
            return self._get_file_batch(batch_files, file_type=file_type, flatten=flatten)

        # otherwise read files in parallel

        # parallel operation, using direct view
        rc = Client()
    
        # use cloudpickle because pickle won't work with plugins
        rc[:].use_cloudpickle()

        # get number of available engines
        num_engines = len(rc)

        # break files into blocks so that we use all available engine
        num_files = len(batch_files)

        # push out parallel jobs, one load per engine
        data = []
        for i in range(0, num_files, num_engines):
            
            # get block of files
            block_files = [os.path.abspath(batch_files[j])
                for j in range(i, i + num_engines) if j < num_files]

            # push out jobs per engine
            async_results = []
            for j in range(len(block_files)):
                async_result = rc[j].apply_async(self._get_file_batch, 
                    [block_files[j]], file_type=file_type, flatten=flatten)
                async_results.append(async_result)

            # wait for results
            rc[:].wait(async_results)

            # get results
            for j in range(len(async_results)):

                # get items in order they were put into queue
                async_result = async_results.pop(0)
                
                # retrieve data from result
                data += async_result.get()

            # clean up ipyparallel
            rc.purge_everything()

        # close ZMQ sockets
        rc.close()

        return data

    # helper function to read in a batch of files
    def _get_file_batch(self, batch_files, file_type=None, flatten=True):
        
        # read all files into list
        data = []
        for i in range(len(batch_files)):
                
            # get data from file
            data_i = self.read_file(batch_files[i], file_type=file_type)

            # do plugin preprocessing  (doesn't do anything if not enabled)
            data_i = self.preprocess(data_i, flatten=flatten)

            # add data to list
            data.append(data_i)

        return data

    # generic writer
    def write_file(self, data, file_out, file_type=None):
        """
        Writes time step data from an ensemble to a file.  File type is
        inferred from extension unless provided.

        Args:
            data (`meshio` mesh): mesh data to be written
            file_out (string): file name for output file
            file_type (string): file extension
        """

        # write out data using meshio
        if isinstance(file_type, str):
            meshio.write(file_out, data, file_format=file_type)

        else:
            self.log.error('Unrecognized output file format for "%s" -- file not written.' %
                file_out)

    # generic file converter
    def convert_file(self, file_in, file_out, file_in_type=None, file_out_type=None):
        """
        Converts from file_in to file_out, where file_in can be a string or
        a buffer.  File types are inferred from extensions unless provided.  Uses
        the meshio library.

        Args:
            file_in (string): name of input file
            file_out (string): name of output file
            file_in_type (string): file input format (regardless of extension)
            file_out_type (string): file output format (extension)
        """

        # read in file data
        mesh = self.read_file(file_in, file_type=file_in_type)

        # write out data
        self.write_file(mesh, file_out, file_type=file_out_type)

    def convert_files(self, file_list, output_dir, output_type, input_type=None):
        """
        Converts a list of files to file of type output_type in output_dir with
        same root name.  Input file types are inferred from extensions, unless type
        is provided.  Output type is also inferred, unless provided.

        Args:
            file_list (list): list of file names to read
            output_dir (string): name of output directory to write files
            output_type (string): extension of file format for output
            input_type (sring): file input type (regardless of extension)

        Returns:
            files_written (list): list of files written using full path
        """

        # keep track of files written
        files_written = []

        # convert files in list
        for file_to_convert in file_list:

            # create output file name
            file_name = os.path.basename(file_to_convert)
            file_root, file_ext = os.path.splitext(file_name)
            file_out = Path(
                os.path.join(output_dir, file_root + "." + output_type)).as_posix()
            
            # convert file
            self.convert_file(file_to_convert, file_out, 
                file_in_type=input_type, file_out_type=output_type)

            files_written.append(file_out)
        
        return files_written

    # input deck reader
    def read_input_deck (self, file_list, file_type=None):
        """
        Reads a file or files which provide the input parameters for a simulation.
        Note that this code must be provided by the plugin.

        Args: 
            file_list (list): list of file names to read (can be a list of one file)
            file_type (string): file type (regardless of extension)
        
        Returns:
            file_data (object): meta data for the simulation
        """

        return None

    # pre-processing specific to simulation
    def preprocess (self, data):
        """
        Performs data pre-processing specific to a simulation.  This code must
        be provided by the plugin, otherwise the data is returned unchanged.  Note
        that this type of pre-processing is assumed to be per file (e.g. per time
        step or per simulation).  Pre-processing that occurs over the entire
        ensemble is done by the algorithm codes (e.g. dimension reduction or 
        proxy models).

        Args: 
            data (object): data to be pre-processed

        Returns:
            data_out (2d array): pre-processed numpy array with simulations per row
        """

        return data

    # expand table by reading file contents
    def expand (self, table, header, file_list, **kwargs):
        """
        Expands a column in a ensemble table by reading the file links and 
        creating files appropriate to the plugin.

        Args:
            table (ensemble Table object): table with column containing file links
            header (string or int): name of column to with file links
            file_list (list): list of files to expand
            **kwargs: additional arguments dependent on plugin
        """

        pass

# factory function to instantiate and initialize a plugin 
# using a module from a file and a list of arguments
def plugin (plugin_name, arg_list=None):
    """
    Factory function to instantiate a plugin module from a file and a list of
    command line arguments.

    Args:
        plugin_name (string): module name (no .py) or file name of module (ending in .py)
        arg_list (list): list containing command line flags and argumetns

    Returns:
        plugin (object), unknown_args (list): plugin as Python namespace,
        list of un-recognized arguments

    :Example:

    .. code-block:: python

        # import and initialize plugin
        plugin, unknown_args = slypi.ensemble.plugin(args.plugin, arg_list)     
    """

    # get plugin name extension
    ext = os.path.splitext(plugin_name)

    # if not a file name, check in plugins directory
    mod = None
    if ext != '.py':

        # go through each module in plugins directory
        for loader, mod_name, is_pkg in pkgutil.iter_modules(
            plugins.__path__, plugins.__name__ + "."):

            # load module from plugins directory
            if mod_name.endswith("." + plugin_name): 
                mod = importlib.import_module(mod_name)
        
    else:

        # load plugin from file
        if os.path.isfile(plugin_name):
            spec = importlib.util.spec_from_file_location(name, plugin_name)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)

    # instantiate new plugin, if found
    if mod is None:
        log.error('Plugin not found, please check spelling and try again.')
        sys.exit(1)
    else:
        plug = mod.Plugin()

    # parse arguments 
    args, arg_list = plug.parse_args(arg_list)

    # check arguments
    plug.check_args(args)

    # initialize plugin variables
    plug.init(args)

    log.debug("Loaded plugin: %s." % plugin_name)

    # return plugin ready to go
    return plug, arg_list



# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module contains ai/ml algorithms for use with slypi.

# S. Martin
# 1/12/2025

# auto-encoder
import torch
import torch.multiprocessing as mp

import time
import warnings

# types of autoencoders
MODEL_TYPES = ["MLP", "var"]

# auto-encoder training defaults
BATCH_SIZE = 64
EPOCHS = 10
LEARNING_RATE = 1e-3
MLP_ARCH = [512,128]
NUM_PROCESSES = 1


class TorchDataset(torch.utils.data.Dataset):
    """
    This class contains a torch dataset constructor that converts a numpy
    array with rows as points to a torch dataset for use with the 
    auto-encoder.

    :Example:

    .. code-block:: python

        # initialize a random dataset, 100 data points in 10 dimensions
        random_data = numpy.random.rand (100,10)

        # convert to a torch dataset
        torch_data = TorchDataset(random_data)

        # put data in a torch dataloader using a batch size of 10
        data_loader = torch.utils.data.DataLoader(torch_data, batch_size=10, shuffle=True)

    """

    # stores a numpy array as a torch tensor
    def __init__(self, X):        
        self.X = torch.from_numpy(X)

    # number of samples
    def __len__(self):
        return self.X.size()[0]

    # get item
    def __getitem__(self, index):
        return self.X[index,:]


class MLP(torch.nn.Module):
    """
    This class contains a basic MLP auto-encoder that can be used for dimension 
    reduction.  The class implements only the pytorch architecture for a model.
    """

    # initialize the network using the input dimension and bottleneck dimension
    def __init__(self, input_dim, bottleneck_dim, arch=MLP_ARCH):
        
        super().__init__()

        # keep track of network architecture
        self.num_inputs = input_dim
        self.num_dim = bottleneck_dim

        # encoder/decoder layers
        self.encoder_arch = [self.num_inputs] + arch
        self.decoder_arch = [self.num_dim] + arch[::-1]
        self.num_layers = len(self.encoder_arch) - 1

        # set up architecture
        self._init_encoder()
        self._init_decoder()

    # define encoder layers (num_inputs -> MLP -> num_dim)
    def _init_encoder(self):

        # append hidden layers
        self.encoder_layers = torch.nn.ModuleList()
        for k in range(self.num_layers):
            self.encoder_layers.append(
                torch.nn.Linear(self.encoder_arch[k], self.encoder_arch[k+1]))

        # bottleneck layer
        self.bottleneck_layer = torch.nn.Linear(self.encoder_arch[-1], self.num_dim)

    # define decoder layers (num_dim -> reverse MLP -> num_inputs)
    def _init_decoder(self):        

        # append hidden layers
        self.decoder_layers = torch.nn.ModuleList()
        for k in range(self.num_layers):
            self.decoder_layers.append(
                torch.nn.Linear(self.decoder_arch[k], self.decoder_arch[k+1]))
        
        # reconstruction layer
        self.reconstruction_layer = torch.nn.Linear(self.decoder_arch[-1], self.num_inputs)

    # execute encoder
    def _encoder(self, data):

        # initial encoder layers
        for layer in self.encoder_layers:
            data = torch.nn.functional.relu(layer(data))

        # bottleneck layer
        bottleneck = self.bottleneck_layer(data)

        return bottleneck

    # execute decoder
    def _decoder(self, data):

        # decoder
        for layer in self.decoder_layers:
            data = torch.nn.functional.relu(layer(data))

        return torch.tanh(self.reconstruction_layer(data))

    # auto-encoder architecture definition for pytorch
    def forward(self, data):

        # encode-decode
        bottleneck = self._encoder(data)
        reconstruction = self._decoder(bottleneck)

        return reconstruction

    # return reduced data using encoder on numpy array
    def encoder(self, data):

        # convert numpy array to torch format
        torch_data = torch.from_numpy(data.astype(np.float32))

        # get bottleneck layer
        bottleneck = self._encoder(torch_data)

        # convert back to numpy array
        return bottleneck.data.numpy()

    # return reconstructed data using decoder on numpy array
    def decoder(self, data):

        # convet numpy array to torch format
        torch_data = torch.from_numpy(data.astype(np.float32))

        # reconstruct data using decoder
        recon = self._decoder(torch_data)

        # convert back to numpy array
        return recon.data.numpy()


# variational autoencoder architecture
class VAE(torch.nn.Module):
    """
    This class contains a basic variational auto-encoder that can be used for dimension 
    reduction.  The class implements only the pytorch architecture for a model.
    """

    # initialize network using input, bottleneck dimension
    def __init__(self, input_dim, bottleneck_dim):

        super().__init__()

        # network architecture
        self.input_dim = input_dim
        self.bottleneck_dim = bottleneck_dim

        # set up architecture
        self._init_encoder()
        self._init_decoder()

    # encoder architecture
    def _init_encoder(self):

        self.fc1 = torch.nn.Linear(self.input_dim, 500)
        self.fc21 = torch.nn.Linear(500, self.bottleneck_dim)
        self.fc22 = torch.nn.Linear(500, self.bottleneck_dim)
    
    # decoder architecture
    def _init_decoder(self):

        self.fc3 = torch.nn.Linear(self.bottleneck_dim, 500)
        self.fc4 = torch.nn.Linear(500, self.input_dim)

    # execute encoder (using torch data)
    def _encode(self, x):

        h1 = torch.nn.functional.relu(self.fc1(x))
        return self.fc21(h1), self.fc22(h1)

    # execute decoder (using torch data)
    def _decode(self, z):
    
        h3 = torch.nn.functional.relu(self.fc3(z))
        return torch.sigmoid(self.fc4(h3))

    # reparameterize for variational layer
    def _reparameterize(self, mu, logvar):
        std = torch.exp(0.5*logvar)
        eps = torch.randn_like(std)
        return mu + eps*std

    # execute auto-encoder (torch data)
    def forward(self, x):
        mu, logvar = self._encode(x)
        z = self._reparameterize(mu, logvar)
        return self._decode(z), mu, logvar

    # return reduced data using encoder on numpy array
    def encoder(self, data):

        # convert numpy array to torch format
        torch_data = torch.from_numpy(data.astype(np.float32))

        # get bottleneck layer (variational)
        mu, logvar = self._encode(torch_data)

        # re-parameterize
        bottleneck = self._reparameterize(mu, logvar)

        # convert back to numpy array
        return bottleneck.data.numpy()

    # return reconstructed data using decoder on numpy array
    def decoder(self, data):

        # convet numpy array to torch format
        torch_data = torch.from_numpy(data.astype(np.float32))

        # reconstruct data using decoder
        recon = self._decode(torch_data)

        # convert back to numpy array
        return recon.data.numpy()


class AutoEncoder(torch.nn.Module):
    """
    This class is a factory class defines and trains a pytorch auto-encoder model.
    The class implements the basic functions expected from the DimensionReduction 
    class, including fit, partial_fit, and transform.
    """

    # initialize auto-encoder with user parameters
    def __init__(self, log, num_dim=NUM_DIM, batch_size=None,
            learning_rate=LEARNING_RATE, epochs=None, num_processes=NUM_PROCESSES,
            model_type='MLP'):

        super().__init__()

        # number of encoder output neurons
        self.num_dim = num_dim

        # set default batch size
        self.batch_size = batch_size
        if self.batch_size is None:
            self.batch_size = BATCH_SIZE

        # learning rate for training
        self.learning_rate = learning_rate

        # set default epochs
        self.epochs = epochs
        if self.epochs is None:
            self.epochs = EPOCHS

        # use multiprocessing
        self.num_processes = num_processes
        if self.num_processes is None:
            self.num_processes = NUM_PROCESSES

        # use calling function log for output
        self.log = log

        # model type for autoencoder
        self.model_type = model_type

        # actual autencoder model has not been initialized
        self.model = None

    # variational reconstruction + KL divergence losses summed over all elements and batch
    def _var_loss(self, recon_x, x, mu, logvar):

        BCE = torch.nn.functional.binary_cross_entropy(recon_x, x, reduction='sum')

        # see Appendix B from VAE paper:
        # Kingma and Welling. Auto-Encoding Variational Bayes. ICLR, 2014
        # https://arxiv.org/abs/1312.6114
        # 0.5 * sum(1 + log(sigma^2) - mu^2 - sigma^2)
        KLD = -0.5 * torch.sum(1 + logvar - mu.pow(2) - logvar.exp())

        return BCE + KLD

    # train auto-encoder (can be called in parallel)
    def _train(self, rank, model, model_type, device, data_loader, 
        optimizer, scheduler, epochs):

        # for MLP use mean square error loss
        if model_type == 'MLP':
            criterion = torch.nn.MSELoss()

        # train auto-encoder
        model.train()
        for epoch in range(epochs):
            for batch_idx, data in enumerate(data_loader):

                # zero gradients
                optimizer.zero_grad()

                # re-construct data & compute loss
                if model_type == 'MLP':
                    recon = model(data.to(device))
                    loss = criterion(recon.to(device), data)
                elif model_type == 'var':
                    recon, mu, logvar = model(data.to(device))
                    loss = self._var_loss(recon.to(device), data, mu, logvar)
                
                # do back-propagation
                loss.backward()
                optimizer.step()

                # using print because we are in a thread
                if batch_idx % 10 == 0:
                    print('Process {} -- Train Epoch: {} [{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                        rank, epoch, batch_idx * len(data), len(data_loader.dataset),
                        100. * batch_idx / len(data_loader), loss.item()))

            # adjust optimization parameters
            scheduler.step(loss.item())

    # define and fit auto-encoder
    def fit(self, data):

        # check batch size
        num_data, num_inputs = data.shape
        if self.batch_size > num_data:
            warnings.warn("Batch size exceeds number of data points -- resetting to " +
                          "number data points.")
            self.batch_size = num_data

        # convert data to torch format
        torch_data = TorchDataset(data.astype(np.float32))
        
        # set up encoder/decoder layers/optimizer
        if self.model is None:

            if self.model_type == 'MLP':
                self.model = MLP(num_inputs, self.num_dim)
            
            elif self.model_type == 'var':
                self.model = VAE(num_inputs, self.num_dim)

            # use Adam optimizer
            self.optimizer = torch.optim.AdamW(self.parameters(),
                                               lr=self.learning_rate,
                                               weight_decay=1e-5)

            # use reduce on stall scheduler
            self.scheduler = torch.optim.lr_scheduler.ReduceLROnPlateau(self.optimizer, patience=50)

        # use CUDA, if available
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        # share model between processes
        self.model.to(device)
        self.model.share_memory()

        # time training 
        t0 = time.time()

        processes = []
        for rank in range(self.num_processes):

            # partition data per process
            data_loader = torch.utils.data.DataLoader(
                dataset=torch_data,
                sampler=torch.utils.data.distributed.DistributedSampler(
                    dataset=torch_data,
                    num_replicas=self.num_processes,
                    rank=rank
                ),
                batch_size=self.batch_size,
            )

            # launch processes
            process = mp.Process(target=self._train, 
                args=(rank, self.model, self.model_type, device, 
                    data_loader, self.optimizer, self.scheduler, self.epochs))
            process.start()
            processes.append(process)

        # collect results
        for process in processes:
            process.join()

        self.log.info("Training Time: %s" % str(timedelta(seconds=time.time()-t0)))

    # perform dimension reduction
    def transform(self, data):
        
        # get reduced data from encoder
        return self.model.encoder(data)

    # inverse transofmr
    def inverse_transform(self, data):

        recon = self.model.decoder(data)

        return recon


class AI:

    # init parser
    def _init_parser(self):

        # auto-encoder parameters

        # model type
        self.parser.add_argument("--model-type", choices=MODEL_TYPES,
            help="Type of auto-encoder.  Options are: {%s}." % ", ".join(MODEL_TYPES))

        # number of epochs
        self.parser.add_argument("--epochs", type=int, help="Number of epochs "
            "to use for training auto-encoder.")

        # batch size
        self.parser.add_argument("--batch-size", type=int, help="Batch size "
            "for training auto-encoder.")

        # MLP architecture
        self.parser.add_argument('--MLP-arch', type=int, nargs='+', help="Integers specifying "
            "size of hidden layers in the MLP.")

        # number of processors
        self.parser.add_argument("--num-processes", type=int, help="Number of processes to "
            "use for training auto-encoder.")
        
    # check parameters for valid values
    def _check_args(self, args):

        # check for specific auto-encoder arguments
        if args.algorithm == "auto-encoder":

            # check epochs >= 1
            if args.epochs is not None:
                if args.epochs < 1:
                    self.log.error("Epochs must be >= 1.")
                    raise ValueError("epochs must be >= 1.")
            
            # check batch-siae >= 1
            if args.batch_size is not None:
                if args.batch_size < 1:
                    self.log.error("Batch size must be >= 1.")
                    raise ValueError("batch size must be >= 1.")

            # check number of process is >= 1
            if args.num_processes is not None:
                if args.num_processes < 1:
                    self.log.error("Number of processes must be >= 1.")
                    raise ValueError("number of prodesses must be >= 1.")

            # check that MLP layer sizes are >= 1
            if args.MLP_arch is not None:
                for layer_size in args.MLP_arch:
                    if layer_size < 1:
                        self.log.error("MLP architecture layer sizes must be >= 1.")
                        raise ValueError("MLP architecture layer sizes must be >= 1.") 
                    
    # set parameters using arguments
    def _set_parms(self, args):

        # auto-encoder arguments
        if args.model_type is None:
            self.model_parms["model_type"] = MODEL_TYPES[0]
        else:
            self.model_parms["model_type"] = args.model_type
        self.model_parms["epochs"] = args.epochs
        self.model_parms["batch_size"] = args.batch_size
        self.model_parms["num_processes"] = args.num_processes


    # add particular algorithm to end of model list
    def _init_algorithm(self, num_dim):

        if self.model_parms["algorithm"] == "auto-encoder":
            self.model.append(AutoEncoder(self.log, num_dim = num_dim,
                                          model_type=self.model_parms["model_type"],
                                          epochs=self.model_parms["epochs"],
                                          batch_size = self.model_parms["batch_size"],
                                          num_processes=self.model_parms["num_processes"]))
            
    # perform incremental dimension reduction
    def partial_fit(self, data, time_step=0):
        """
        Train an incremental model using samples.

        Args:
            data (array): data with points as rows
            time_step (int): model time step
        """

        if self.model_parms["algorithm"] == "auto-encoder":
            self.model[time_step].fit(data)
        
        


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC . Under the terms of Contract
# DE-NA0003525 with National Technology and Engineering Solutions of Sandia, LLC, the U.S. Government
# retains certain rights in this software.

# This script reads a local videoswarm format consisting of a .csv, .xcoords, .ycoords,
# and .trajectories files.  These files are the same files that are selected using the "local"
# option in the videoswarm wizard.

import os
import urllib.parse as urlparse
import numpy as np

import slypi

# slycat parsers
import slypi.pandas_util
import slypi.vs.vs_parse_files as vs_parse

# parse command line arguments
def parser():

    parser = slypi.ArgumentParser(description=
        "Create Slycat VideoSwarm model from local format (.csv, .xcoords, .ycoords, .trajectories) files.")

    # required named arguments
    required = parser.add_argument_group('required arguments')

    # input files
    required.add_argument("--csv-file", help="Input .csv file.", required=True)
    required.add_argument("--xcoords-file", help="Input .xcoords file.", required=True)
    required.add_argument("--ycoords-file", help="Input .ycoords file.", required=True)
    required.add_argument("--traj-file", help="Input .trajectories file.", required=True)

    # video column
    required.add_argument("--video-column", help="Header name of video column.", required=True)

    # options to modify video column link
    parser.add_argument("--video-hostname", default=None, 
        help="Optionally override the hostname for the video column.")
    parser.add_argument("--strip", type=int, default=None, 
        help="Optionally strip N prefix directories from the video URLs.")

    # model and project names/descriptions
    parser.add_argument("--marking", default="mna", 
        help="Marking type.  Default: %(default)s")
    parser.add_argument("--model-description", default="", 
        help="New model description.  Default: %(default)s")
    parser.add_argument("--model-name", default="SlyPI VS Model", 
        help="New model name.  Default: %(default)s")
    parser.add_argument("--project-description", default="", 
        help="New project description.  Default: %(default)s")
    parser.add_argument("--project-name", required=True, 
        help="New project name.")

    return parser

# logging is just printing to the screen
def log (msg):
    print(msg)

# create VS model, arguments include the command line connection parameters
# and project/model information
def upload_model (arguments, csv_data, xcoords_data, ycoords_data, traj_data, links, log):

    # setup a connection to the Slycat Web Server.
    connection = slypi.connect(arguments)

    # create a new project to contain our model.
    pid = connection.find_or_create_project(arguments.project_name, arguments.project_description)

    # create the new, empty model.
    mid = connection.post_project_models(pid, "VS", 
        arguments.model_name, arguments.marking, arguments.model_description)

    # upload model table as movies.meta
    log('Uploading table data ...')
    connection.put_model_arrayset(mid, "movies.meta")
    connection.put_model_arrayset_array(mid, "movies.meta", 0, 
        csv_data['dimensions'], csv_data['attributes'])
    for index in range(len(csv_data['data'])):
        values = [np.asarray(csv_data['data'][index])]
        connection.put_model_arrayset_data(mid, "movies.meta", "0/%s/..." % index, values)

    # upload movie links as movies.links
    attributes = [dict(name="value", type="string")]
    dimensions = [dict(name="row", end=int(len(links)))]
    connection.put_model_arrayset(mid, "movies.links")
    connection.put_model_arrayset_array(mid, "movies.links", 0, dimensions, attributes)
    connection.put_model_arrayset_data(mid, "movies.links", "0/.../...", [np.asarray(links)])

    # upload x-coords
    log('Uploading coordinates ...')
    connection.put_model_arrayset(mid, "movies.xcoords")
    connection.put_model_arrayset_array(mid, "movies.xcoords", 0,
        xcoords_data['dimensions'], xcoords_data['attributes'])
    connection.put_model_arrayset_data(mid, "movies.xcoords", "0/.../...", [xcoords_data['data']])

    # upload y-coords
    connection.put_model_arrayset(mid, "movies.ycoords")
    connection.put_model_arrayset_array(mid, "movies.ycoords", 0,
        ycoords_data['dimensions'], ycoords_data['attributes'])
    connection.put_model_arrayset_data(mid, "movies.ycoords", "0/.../...", [ycoords_data['data']])

    # upload trajectories
    log('Uploading trajectories ...')
    connection.put_model_arrayset(mid, "movies.trajectories")
    connection.put_model_arrayset_array(mid, "movies.trajectories", 0,
        traj_data['dimensions'], traj_data['attributes'])
    connection.put_model_arrayset_data(mid, "movies.trajectories", "0/.../...", [traj_data['data']])

    # note that this model was loaded via push script
    connection.put_model_parameter(mid, "vs-loading-parms", ["Uploaded"])
    
    # Signal that we're done uploading data to the model.  This lets Slycat Web
    # Server know that it can start computation.
    connection.post_model_finish(mid)

    # Wait until the model is ready.
    connection.join_model(mid)

    return mid

# check arguments and create model, returns URL to model
def create_model (arguments, log):

    # check that input files exist
    for file in [arguments.csv_file, arguments.xcoords_file, 
                 arguments.ycoords_file, arguments.traj_file]:
        if not os.path.isfile(file):
            log("Input file " + file + " does not exist.")
            raise ValueError("Input file " + file + " does not exist.")
    
    # load csv file
    log('Reading .csv file ...')
    attributes, dimensions, data, csv_read_error = \
        slypi.pandas_util.parse_file(arguments.csv_file, file_name=True)

    # check for warnings/errors
    for i in range(len(csv_read_error)):
        if csv_read_error[i]["type"] == "warning":
            log("Warning: " + csv_read_error[i]["message"])
        else:
            log("Error: " + csv_read_error[i]["message"])
            raise ValueError(csv_read_error[i]["message"])

    # get column names/types
    column_names = [attribute["name"] for attribute in attributes]
    column_types = [attribute["type"] for attribute in attributes]

    # check that video column input is in headers
    if arguments.video_column not in column_names:
        log('Video column "' + arguments.video_column + '" not found in input file.')
        raise ValueError('Video column "' + arguments.video_column + '" not found in input file.')

    # get video index in csv
    video_index = column_names.index(arguments.video_column)

    # check that video column has strings
    if column_types[video_index] != "string":
        log('Video column "' + arguments.video_column + '" does not contain strings.')
        raise ValueError('Video column "' + arguments.video_column + '" does not contain strings.')

    # replace video URL hostnames, if requested
    if arguments.video_hostname is not None:

        # replace file://cluster with file://hostname
        modified = []
        for uri in data[video_index]:
            uri = urlparse.urlparse(uri)
            path = uri.path
            uri = "file://%s%s" % (arguments.video_hostname, path)
            modified.append(uri)
            data[video_index] = modified

    # strip prefix directories, if requested
    if arguments.strip is not None:

            # strip prefixes
            modified = []
            for uri in data[video_index]:
                uri = urlparse.urlparse(uri)
                hostname = uri.netloc
                path = '/'.join(uri.path.split('/')[arguments.strip + 1:])
                uri = "file://%s/%s" % (hostname, path)
                modified.append(uri)
                data[video_index] = modified

    # consolidate csv file
    csv_data = {'attributes': attributes, 'dimensions': dimensions, 'data': data}

    # finalize video links
    video_links = data[video_index]

    # load .xcoords file as a string
    log('Reading .xcoords file ...')
    with open(arguments.xcoords_file, 'r') as file:
        mat_file = file.read()

    # parse .xcoords file
    attributes, dimensions, data = vs_parse.parse_mat_file(mat_file)
    xcoords_data = {'attributes': attributes, 'dimensions': dimensions, 'data': data}

    # check that .xccords file has the same number of points as the csv file
    if dimensions[1]['end'] != csv_data['dimensions'][0]['end']:
        log('Different number of points in .xcoords file and .csv file.')
        raise ValueError('Different number of points in .xcoords file and .csv file.')
    
    # load .ycoords file as a string
    log('Reading .yxcoords file ...')
    with open(arguments.ycoords_file, 'r') as file:
        mat_file = file.read()

    # parse .ycoords file
    attributes, dimensions, data = vs_parse.parse_mat_file(mat_file)
    ycoords_data = {'attributes': attributes, 'dimensions': dimensions, 'data': data}
    
    # check that .yccords file has the same number of points as the csv file
    if dimensions[1]['end'] != csv_data['dimensions'][0]['end']:
        log('Different number of points in .xcoords file and .csv file.')
        raise ValueError('Different number of points in .xcoords file and .csv file.')
    
    # check that .xcoords and .ycoords files have the same number of time steps
    if dimensions[0]['end'] != xcoords_data['dimensions'][0]['end']:
        log('Different number of time steps in .xcoords file and .ycoords file.')
        raise ValueError('Different number of time steps in .xcoords file and .ycoords file.')

    # load .trajectories file
    log('Reading .trajectories file ...')
    with open(arguments.traj_file, 'r') as file:
        mat_file = file.read()

    # parse .trajectories file
    attributes, dimensions, data = vs_parse.parse_mat_file(mat_file)
    traj_data = {'attributes': attributes, 'dimensions': dimensions, 'data': data}

    # check number of trajectories
    if dimensions[0]['end'] != xcoords_data['dimensions'][1]['end']+1:
        log('Different number of trajectories than points in coordinates files.')
        raise ValueError('Different number of trajectories than points in coordinates files.')

    # check number of time steps in trajectories
    if dimensions[1]['end'] != xcoords_data['dimensions'][0]['end']:
        log('Different number of time steps in coordinate files and .trajectories file.')
        raise ValueError('Different number of time steps in coordinate files and .trajectories file.')

    # echo inputs to user
    log('Input .csv file: ' + arguments.csv_file)
    log('Input .xcoords file: ' + arguments.xcoords_file)
    log('Input .ycoords file: ' + arguments.ycoords_file)
    log('Input .trajectories file: ' + arguments.traj_file)
    log('Input video column name: ' + arguments.video_column)
    if arguments.video_hostname is not None:
        log('Video hostname: ' + str(arguments.video_hostname))
    if arguments.strip is not None:
        log('Stripped ' + str(arguments.strip) + ' prefixes from video column.')
    if arguments.strip is not None or arguments.video_hostname is not None:
        log('First new video link: ' + csv_data['data'][video_index][0])

    # report on number of points and time steps
    log('Found ' + str(csv_data['dimensions'][0]['end']) + ' data points and ' +
        str(xcoords_data['dimensions'][0]['end']) + ' time steps.')

    # upload model
    mid = upload_model (arguments, csv_data, xcoords_data, ycoords_data, traj_data, video_links, log)

    # supply direct link to model
    host = arguments.host
    if arguments.port:
        host = host + ":" + arguments.port
    url = host + "/models/" + mid
    log("Your new model is located at %s" % url)
    log('***** VideoSwarm Model Successfully Created *****')

    # return url
    return url

# main entry point
def main():

    # set up argument parser
    vs_parser = parser()  

    # get user arguments
    arguments = vs_parser.parse_args()

    # check arguments and create model
    url = create_model(arguments, log)

# command line version
if __name__ == "__main__":

    main()



# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.



# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module contains all of the available algorithms for
# performing dimension reduction using the slypi ensemble tools.

# This code serves mainly as a wrapper for algorithms available
# from other libraries, such as sklearn.

# S. Martin
# 2/17/2021


# standard library imports

# command line arguments
import argparse

# logging and error handling
import logging
import warnings

# model save/load
import joblib

# 3rd party imports

# matrix manipulations
import numpy as np

# pre-processing
import sklearn.preprocessing as preprocessing

# dimension reduction
import sklearn.decomposition as decomposition
import sklearn.manifold as manifold

# umap
import umap

# comparing runing times (for debugging)
import time
from datetime import timedelta

# list of available preprocessing and dimension reduction algorithms
PREPROCESS = ["standard", "minmax"]
ALGORITHMS = ["PCA", "incremental-PCA", "Isomap", 
              "tSNE", "Umap"]
INCREMENTAL = [False, True, False, 
               False, True, False, False]
INVERSE = [True, True, False,
           False, True, False, True]

# number of default dimensions for reduction
NUM_DIM = 2

class DimensionReduction:
    """
    This class contains wrappers for dimension reduction algorithms in sci-kit
    learn for use with the slypi ensemble tools.  It includes it's 
    own parser to specify algorithms and algorithm parameters.

    Args:
        arg_list (list): list of arguments to specific to reduction
        model_file (string): name of model file containing reduction

    :Example:

    .. code-block:: python

        # get parser and reduction algorithm code
        import slypi.ensemble as ensemble
        import slypi.algorithms.reduction as algorithms

        # parse command line
        my_parser = ensemble.ArgumentParser()

        # parse command line and start logger
        args, arg_list = my_parser.parse_args()

        # set up dimension reduction algorithm using command line arguments
        algorithm = algorithms.DimensionReduction(arg_list=arg_list)

        # use time_align to use a time-aligned model, where time_align 
        # specifies the number of dimension to use per time step
        time_aligned_algorithm = algorithms.DimensionReduction(time_align=10)

        # set up data in variable X, data points per row

        # do dimension reduction (ala sklearn)
        algorithm.fit(X)

        # reduced data to lower dimension
        reduced_data = algorithm.transform(X)

    """

    # set up dimension reduction algorithm specific arguments
    def __init__(self, arg_list=None, model_file=None):

        # create parser
        self._init_parser()

        # start logger
        self.log = logging.getLogger("slypi.ensemble.algorithms.reduction")
        self.log.debug("Started dimension reduction.")

        # parse known arguments, keep unknown arguments
        args, arg_list = self.parser.parse_known_args(arg_list)
        self.unknown_arglist = arg_list

        # check for model inputs passed as arguments
        parms_present = [True for key in vars(args).keys() \
                if vars(args)[key] is not None]
        parms_present = len(parms_present) > 0

        # check that arg_list or model_file is provided
        if not parms_present and model_file is None:
            self.log.error("Must provide either model input file or model arguments to " +
                "perform dimension reduction.")
            raise ValueError("must provide either model input file or model arguments to " +
                "perform dimension reduction.")

        # check that both arg_list and model_file are not provided
        if parms_present and model_file is not None:
            self.log.error("Use either a model input file or model arguments, but not both.")
            raise ValueError("use either a model input file or model arguments, but not both.")

        # initialize model, use list for time aligned case
        self.model = []
        self.pre_process = []

        # initialize model parameters
        self.model_parms = {}
        self.model_parms["num_dim"] = None
        self.model_parms["algorithm"] = None
        self.model_parms["pre_process"] = None

        # parse & check arguments, if given
        if parms_present:
            
            # check arguments
            self._check_args(args)

            # init model
            self._set_parms(args)
            self._init_model()

        # load a model, if requested
        if model_file is not None:
            self.load(model_file)

    # init parser
    def _init_parser(self):

        # includes parser for dimension reduction options
        description = "Dimension reduction support for the slypi ensemble tools."
        self.parser = argparse.ArgumentParser(description=description)

        # type of algorithm
        self.parser.add_argument("--algorithm", choices=ALGORITHMS, help=
            "Dimension reduction algorithm to apply.  Options are: {%s}." % 
            ", ".join(ALGORITHMS))

        # type of pre-processing
        self.parser.add_argument("--pre-process", choices=PREPROCESS, help=
            "Preprocessing for dimension reduction.  Options are: {%s}." %
            ", ".join(PREPROCESS))

        # number of dimensions (required for any dimension reduction algorithm)
        self.parser.add_argument("--num-dim", type=int, help="Number of desired "
            "dimensions in reduction.")

        # time aligned option
        self.parser.add_argument("--time-align", type=int, help="Train reduction model " 
            "per time step to given dimension then align using Kabsch algorithm.")

        # PCA parameters
        self.parser.add_argument('--whiten', action="store_true", default=None,
            help="Whiten before PCA.")

    # check parameters for valid values
    def _check_args(self, args):

        # check if algorithm is present
        if args.algorithm is None:
            self.log.error("Algorithm choice is required for dimension reduction.  Please use " +
                           "--algorithm and try again.")
            raise ValueError("algorithm choice is required for dimension reduction.  Please use " +
                             "--algorithm or a model input file and try again.")

        # check if number of dimensions exists
        if args.num_dim is None:
            self.log.error("Number of dimensions in reduction must be provided.")
            raise ValueError("number of dimensions in reduction must be provided.")
        
        # check if number of dimensions >= 2
        else:
            if args.num_dim < 2:
                self.log.error("Number of dimensions in reduction must be integer >= 2.")
                raise ValueError("number of dimensions in reduction must be integer >= 2.")

        # check to time aligned >= 2
        if args.time_align is not None:
            if args.time_align < 2:
                self.log.error("Time aligned dimnension must be integer >= 2.")
                raise ValueError("time algined dimension must be integer > = 2.")

    # set parameters using arguments
    def _set_parms(self, args):

        # general arguments
        self.model_parms["num_dim"] = args.num_dim
        self.model_parms["algorithm"] = args.algorithm
        self.model_parms["pre_process"] = args.pre_process

        # PCA arguments (must be True or False)
        if args.whiten is None:
            self.model_parms["whiten"] = False
        else:
            self.model_parms["whiten"] = args.whiten

        # time alignment arguments
        self.model_parms["time_align_dim"] = args.time_align

    # add particular algorithm to end of model list
    def _init_algorithm(self, num_dim):

        # init pre-processing
        if self.model_parms["pre_process"] == "standard":
            self.pre_process.append(preprocessing.StandardScaler())

        elif self.model_parms["pre_process"] == "minmax":
            self.pre_process.append(preprocessing.MinMaxScaler())

        # init algorithm
        if self.model_parms["algorithm"] == "PCA":
            self.model.append(decomposition.PCA(n_components = num_dim, 
            whiten=self.model_parms["whiten"]))
        
        elif self.model_parms["algorithm"] == "incremental-PCA":
            self.model.append(decomposition.IncrementalPCA(n_components = num_dim,
                whiten=self.model_parms["whiten"]))

        elif self.model_parms["algorithm"] == "Isomap":
            self.model.append(manifold.Isomap(n_components = num_dim))
        
        elif self.model_parms["algorithm"] == "tSNE":
            self.model.append(manifold.TSNE(n_components = num_dim))
                            
        elif self.model_parms["algorithm"] == "Umap":
            self.model.append(umap.UMAP(n_components = num_dim))

    # init model according to arguments
    def _init_model(self):

        # set time aligned rotation matrices to empty list by default
        self.align_rot_mats = []

        # for time aligned models we initialize the model when fit is called
        if self.model_parms["time_align_dim"] is None:
            
            # otherwise we add to the first item in model list
            self._init_algorithm(self.model_parms["num_dim"])

    # get any unrecognized arguments
    def unknown_args(self):
        return self.unknown_arglist

    # check if algorithm has partial fit
    def is_incremental(self):
        """
        Test if the user selected an incremental algorithm.

        Returns:
            is_incremental (bool): True if algorithm can be used in batch mode
        """

        # find algorithm in list of algorithms
        alg_ind = ALGORITHMS.index(self.model_parms["algorithm"])

        return INCREMENTAL[alg_ind]

    # check if algorithm can use multi-processing
    def has_inverse(self):
        """
        Check if user selected algorithm has an inverse.

        Returns:
            has_inverse (bool): True if algorithm has an inverse
        """

        # find algorithm in list of algorithms
        alg_ind = ALGORITHMS.index(self.model_parms["algorithm"])

        return INVERSE[alg_ind]

    # return goal number of dimensions
    def num_dim(self):
        """
        Get desired number of dimensions in reduction.

        Returns:
            num_dim (int): number of dimensions for desired reduction
        """

        return self.model_parms["num_dim"]

    # return number of dimensions to use for alignment
    def time_align_dim(self):
        """
        Get number of dimensions to use for time alignment.

        Returns:
            time_align (int): number of time alignment dimension
        """

        return self.model_parms["time_align_dim"]
    
    # perform dimension reduction 
    def fit(self, data, time_step=0):
        """
        Train dimension reduction model using samples.

        Args:
            data (array): data with points as rows
            time_step (int): train model at given time step 
        """
        
        # for time aligned models we initialize and fit model
        if self.model_parms["time_align_dim"] is not None:
            
            # check that time_step is in list or at end
            if time_step > len(self.model):
                self.log.error("Model does not exist and cannot be added.")
                raise IndexError("Model does not exist and cannot be added.")

            # add model to list
            if time_step == len(self.model):
                self._init_algorithm(self.model_parms["time_align_dim"])

        # most models in sklearn have a fit function 
        # but for tSNE we use fit_transform instead
        if self.model_parms["algorithm"] == "PCA" or \
            self.model_parms["algorithm"] == "incremental-PCA" or \
            self.model_parms["algorithm"] == "Isomap" or \
            self.model_parms["algorithm"] == "Umap":

            # check for pre-processing
            if self.model_parms["pre_process"] is not None:
                data = self.pre_process[time_step].fit_transform(data)

            self.model[time_step].fit(data)
    
    # perform incremental dimension reduction
    def partial_fit(self, data, time_step=0):
        """
        Train an incremental model using samples.

        Args:
            data (array): data with points as rows
            time_step (int): model time step
        """

        # for time aligne3d models we initialize and do partial fit
        if self.model_parms["time_align_dim"] is not None:

            # check that time_step is in list or at end
            if time_step > len(self.model):
                self.log.error("Model does not exist and cannot be added.")
                raise IndexError("Model does not exist and cannot be added.")

            # add model to list, if at new time step 
            # (otherwise should already be initialized)
            if time_step == len(self.model):
                self._init_algorithm(self.model_parms["time_align_dim"])

        # check for pre-processing
        if self.model_parms["pre_process"] is not None:
            self.pre_process[time_step].partial_fit(data)
            data = self.pre_process[time_step].transform(data)

        if self.model_parms["algorithm"] == "incremental-PCA":

            # make sure we have enough samples
            if data.shape[0] < self.model_parms["num_dim"]:
                self.log.warn("Need more samples to perform partial fit, ignoring batch.")
            else:
                self.model[time_step].partial_fit(data)

    # transform data using model
    def transform(self, data, time_step=0):
        """
        Transform data to lower dimensional representation.

        Args:
            data (numpy array): data with points as rows
            time_step (int): transform data at given time step

        Returns:
            data (array): reduced data
        """

        # check that data is 2D
        if len(data.shape) == 1:
            data = data.reshape(1, -1)

        # check for pre-processing
        if self.model_parms["pre_process"] is not None:
            data = self.pre_process[time_step].transform(data)

        # for tSNE do fit_transform
        if self.model_parms["algorithm"] == "tSNE":
            data = self.model[time_step].fit_transform(data)

        # other models have transform
        elif self.model is not None:
            data = self.model[time_step].transform(data)

        return data
    
    # inverse transform from data
    def inverse_transform(self, data, time_step=0):

        # inverse transform of latent data
        data = self.model[time_step].inverse_transform(data)

        # check for post-processing
        if self.model_parms["pre_process"] is not None:
            data = self.pre_process[time_step].inverse_transform(data)

        return data

    # time align reduced data using Kabsch algorithm
    def time_align(self, data, compute_rotations=True):
        """
        Time align reduced data using the Kabsch algorithm.  Expects the incoming
        dimension to be time_align_dim and truncates the dimension to num_dim
        after alignment.

        Args: data (list of array): list of data matrices of shape (sim, dim)
              compute_rotations (boolean): False to use existing rotation matrices

        Returns:
            aligned_data (list of array): list of data with shape (sim, reduced dim)
        """

        num_time = len(data)
        
        # align from last to first frame
        aligned_data = []
        align_rot_mats = []
        for i in reversed(range(num_time)):

            # rotate to previous coordinates
            if i == num_time-1:
                old_coords = data[i]
                align_rot_mats.insert(0, np.identity(data[i].shape[1]))

            else:

                # compute new rotations
                if compute_rotations:

                    # do Kabsch algorithm
                    A = data[i].transpose().dot(old_coords)
                    U, S, V = np.linalg.svd(A)
                    rot_mat = (V.transpose()).dot(U.transpose())

                # or use existing rotation matrices
                else:
                    rot_mat = self.align_rot_mats[i]

                # rotate to get new coordinates
                data[i] = data[i].dot(rot_mat.transpose())

                # update old coords
                old_coords = data[i]

                # save rotation matrices as part of model
                align_rot_mats.insert(0, rot_mat)

            # collect aligned coodinates
            aligned_data.insert(0, data[i][:,0:self.model_parms["num_dim"]])

        # save rotation matrices to model
        self.align_rot_mats = align_rot_mats

        return aligned_data
    
    # return percent variance captured per dimension per model
    # if a given algorithm doesn't have this information, return []
    def data_explained(self):
        """
        Returns the percent of information captured per dimension
        per model.  For PCA this would be the explained variance
        ratio.  If a model doesn't compute this information, an
        empty list is returned.

        Returns:
            data_explained (list): list of vectors of percent captured
        """

        variance_ratio_per_model = []
        if self.model_parms['algorithm'] == 'PCA' or \
            self.model_parms['algorithm'] == 'incremental-PCA':

            # collect results per model
            for i in range(len(self.model)):
                variance_ratio_per_model.append(np.cumsum(100 * 
                    self.model[i].explained_variance_ratio_))
        
        return variance_ratio_per_model

    # save dimension reduction model
    def save(self, file_out):
        """
        Saves a dimension reduction model to a .pkl file.

        Args:
            file_out (string): file name to save file
        """

        joblib.dump([self.model, self.pre_process, self.model_parms, self.align_rot_mats], 
            file_out, compress=True)

        self.log.info("Saved model file to %s." % file_out)
        
    # load dimension reduction model
    def load(self, file_in):
        """
        Loads a dimension reduction model from a .pkl file.

        Args:
            file_in (string): file name with saved model
        """

        # load model and parameters
        self.model, self.pre_process, self.model_parms, self.align_rot_mats = joblib.load(file_in)

        self.log.info("Loaded model file from %s." % file_in)


# if called from the command line display algorithm specific command line arguments
if __name__ == "__main__":

    # show options on command line if requested
    algorithms = DimensionReduction()
    algorithms.parser._parse_args()
    
    # anchor multi-processing to this module
    mp.set_start_method('spawn')


# Copyright 2013 National Technology & Engineering Solutions of Sandia, LLC (NTESS). 
# Under the terms of Contract DE-NA0003525 with NTESS, the U.S. Government 
# retains certain rights in this software.
#
# This code does the actual parsing of csv files and matrix files
# for VideoSwarm.
#
# S. Martin
# 5/20/2018

# import cherrypy
import csv
import numpy

# this version can handle either a matrix of floats, or a vector of strings
def parse_mat_file(file):

    """
    parses out a csv file into numpy array by column (data), the dimension meta data(dimensions),
    and sets attributes (attributes)
    :param file: csv file to be parsed
    :returns: attributes, dimensions, data
    """

    # cherrypy.log.error("Started VS generic matrix parser.")

    # parse file using comma delimiter
    rows = [row for row in csv.reader(file.splitlines(), delimiter=",", doublequote=True,
            escapechar=None, quotechar='"', quoting=csv.QUOTE_MINIMAL, skipinitialspace=True)]

    # fill a numpy matrix with matrix from file (assumes floats, uses strings otherwise)
    data = numpy.zeros((len(rows[0:]), len(rows[0])))
    string_data = []
    is_float = True
    for j in range(len(rows[0:])):
        try:
            data[j,:] = numpy.array([float(name) for name in rows[j]])
        except:
            is_float = False
            string_data.append([name for name in rows[j]])

    # set dimensions according to what was in the file
    dimensions = [dict(name="row", end=int(data.shape[0])),
        dict(name="column", end=int(data.shape[1]))]

    # attributes are the same for matrices and vectors
    if is_float:
        attributes = [dict(name="value", type="float64")]
    else:
        attributes = [dict(name="value", type="string")]
        data = string_data

    # for debugging:
    # cherrypy.log.error (str(data))
    # cherrypy.log.error (str(dimensions))
    # cherrypy.log.error (str(attributes))

    return attributes, dimensions, data


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This is a plugin that provides support for converting time-aligned
# dimension reduction to VideoSwarm format.  To be called using
# the --expand option from table.py.  When calling from table.py use
# a dummy value for --csv-out.

# S. Martin
# 3/23/2021


# standard library imports
import csv
import os
from pathlib import Path

# 3rd party imports
import numpy as np

# image manipulation
import imageio
from PIL import Image

# auto-correlation
try:
    import pymks
except ImportError:
    pymks = None

# local imports
import slypi.ensemble as ensemble


# functions for specific operations
# class must be named Plugin
class Plugin(ensemble.PluginTemplate):

    # initialize command line argument parser
    def __init__(self):

        # describe plugin
        description = "The videoswarm plugin provides the ability to convert " + \
            "time-aligned dimension reduction algorithms to Slycat VideoSwarm " + \
            "format."

        # set up parser
        super().__init__(description=description)

    # add any extra command line arguments
    def add_args(self):

        self.parser.add_argument("--remove-expand-col", action="store_true", default=False,
            help="Remove the expanded column when writing out videoswarm files.")
        
        # arguments to specify time scale
        self.parser.add_argument("--video-fps", type=float, default=25,
            help="Video frames per second, must be > 0, defaults to 25.")

        self.parser.add_argument("--binary", action="store_true", help="Converts field "
            "variable to binary by clipping anything less than 0 to 0 and anyting "
            "greater than 0 to 1.")

        self.parser.add_argument("--auto-correlate", action="store_true", help="Performs "
            "auto-correlation as a pre-processing for dimension reduction (note this option "
            "requires the --binary flag to be used).")

        self.parser.add_argument("--scale", action="store_true", help="Scale values to [-0.5,0.5].")

    # check arguments, error if incorrect input
    def check_args(self, args):

        if args.video_fps is not None:
            if args.video_fps <= 0:
                self.log.error("Video frames per second must be > 0.")
                raise ValueError("video frames per second must be > 0 (see " +
                                 "slypi.ensemble.plugins.vs --help).")

        # check that binary is active if auto-correlate is selected
        if args.auto_correlate:

            # check if pymks is available
            if not pymks:
                self.log.error("The pymks module has not been installed, " + 
                                 "cannot perform auto-correlation. " +
                                 "Use pip install slypi[auto].")
                raise ValueError("The pymks module has not been installed, " + 
                                 "cannot perform auto-correlation. " +
                                 "Use pip install slypi[auto].")
            
            # check that binary flag is also present
            if not args.binary:
                self.log.error("Auto-correlation requires binary input, please use --binary " +
                               "and try again.")
                raise ValueError("auto-correlation requires binary flag (see " +
                                 "slypi.ensemble.plugins.vs --help.")

    # initialize any local variables from command line arguments
    def init(self, args):

        # save args for later use
        self.args = args

    # over-ride file read, accept .jpg or .rd.npy files
    def read_file(self, file_in, file_type=None):

        # is it an rd.npy file?
        if file_in.endswith('.rd.npy'):

            # read npy file
            try:
                data = np.load(file_in)
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("could not read " + file_in + " as a .npy file.")

        # is it .npz?
        elif file_in.endswith('.npz'):

            # read npz file
            try:
                data = np.load(file_in)
                data = data['arr_0']
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("could not read " + file_in + " as a .npy file.")

        elif file_in.endswith('.jpg'):

            # read .jpg file and reshape into vector
            data = imageio.imread(file_in)

        elif file_in.endswith('.png'):

            try:
                img = Image.open(file_in)
                data = np.array(img)
            except ValueError:
                self.log.error("Could not read " + file_in + "as a .png file.")
                raise ValueError("Could not read " + file_in + " as a .png file.")

        else:
            self.log.error("The videoswarm plugin accepts only .png, .jpg or .rd.npy files.")
            raise TypeError("videoswarm plugin accepts only .png, .jpg or .rd.npy files.")
        
        self.log.info("Read file " + file_in + ".")

        return data

    # write out .npy format
    def write_file(self, data, file_out, file_type=None):

        # infer file type, if not provided
        if file_type is None:
            if file_out.endswith('.npy'):
                file_type = 'npy'

        # check for npy file output (data is matrix format)
        if file_type == 'npy':
            np.save(file_out, data)
            
        else:

            # default to meshio output
            super().write_file(data, file_out, file_type=file_type)

    # convert from 3 channel image to numpy 
    def _convert_img (self, data):

        if len(data.shape) == 3:

            # convert to black and white image then numpy array
            img = Image.fromarray(data).convert('L')
            data = np.array(img)

        return data

    # scale values between [0,1]
    def _scale(self, data):

        min = data.min()
        max = data.max()

        data = (data - min)/(max - min) - 0.5

        return data

    # make binary version of matrix
    def _binary(self, data):

        data[data < 0] = 0
        data[data > 0] = 1

        return data

    # perform pre-processing on jpg data (n,n,3)
    def preprocess (self, data, flatten=True):
        
        # if image, convert to matrix using jet map
        data = self._convert_img(data)

        # scale data, if requested
        if self.args.scale:
            data = self._scale(data)

        # convert to binary, if requested
        if self.args.binary:
            data = self._binary(data)

        # do auto-correlation, if requested
        if self.args.auto_correlate:
            
            # set up hat basis for pymks
            p_basis = pymks.bases.PrimitiveBasis(n_states=2)

            # check if this data is 2D
            if len(data.shape) == 2:

                # keep rows and columns
                num_rows, num_cols = data.shape

                # add extra dimension for 2D
                data = np.expand_dims(data, axis=0)
            
            # do auto-correlation with pymks (use binary data)
            space_stats = pymks.stats.correlate(data, p_basis, 
                                                periodic_axes=(0, 1), 
                                                correlations=[(0, 0)])

            # re-shape each time step into vector
            space_stats_vec = np.reshape(space_stats,[space_stats.shape[0],
                                         space_stats.shape[1] * space_stats.shape[2],
                                         space_stats.shape[-1]])
            data_vec = np.squeeze(space_stats_vec)

            # re-shape to 2D if we don't want flattened data
            if not flatten:

                data_vec = np.reshape(data_vec, (num_rows, num_cols))
        
        else:

            # default behaviour is to reshape matrix into one simulation per row
            if flatten:

                if len(data.shape) == 2:
                    data_vec = np.reshape(data, [data.shape[0] * data.shape[1]])

                else:
                    data_vec = np.reshape(data, [data.shape[0], 
                                        data.shape[1] * data.shape[2]])

            # otherwise, keep as matrix
            else:
                data_vec = data
                
        # make sure there are no singleton dimensions
        data_vec = np.squeeze(data_vec)
        
        return data_vec

    # over-riding expand to generate videoswarm input files
    def expand(self, table, header, file_list, **kwargs):
        """
        Outputs the videoswarm input files from a set of .rd.npy files.

        Optional arguments include (see also table.py options):
            output_dir -- output diredtory
            csv_out -- csv file to output
            csv_no_index -- do not include pandas index in movies.csv
            csv_index_header -- use this header name for index column
            csv_headers -- output only these columns to movies.csv

        Other arguments passed via kwargs are ignored.
        """

        # decode extra arguments
        output_dir = kwargs.get("output_dir", "")
        csv_no_index = kwargs.get("csv_no_index", False)
        csv_out = kwargs.get("csv_out", None)
        csv_index_header = kwargs.get("csv_index_header", None)
        csv_headers = kwargs.get("csv_headers", None)

        # videoswarm input files are of the form:

        # movies.csv -- .csv metadata file with column containing file links
        # movies.trajectories -- .csv file containing one row of time
        #                         points followed by x-coordinates over simulations
        # movies.xcoords -- .csv file containing one column per
        #                    simulation of x-coordinates over time
        # movies.ycoords -- .csv file containing one column per 
        #                    simulation of y-coordinates over time

        # coordinates are scaled to lie in [0,1]^2

        # file_list is in order of table, expecting .rd.npy files
        xcoords = []
        ycoords = []
        for file_name in file_list:
            
            # read file
            data = self.read_file(file_name)

            # check that data is at least two dimensional
            if data.shape[1] < 2:
                self.log.error("Data in " + file_name + " contains less than " +
                               "two dimensions.")
                raise ValueError("reduced dimension data must have at least two dimensions.")
            
            # collate data (each .rd.npy file contains time steps 
            # per simulation as rows and reduced data as columns)
            xcoords.append(data[:,0])
            ycoords.append(data[:,1])

        # transpose xcoords, ycoords 
        xcoords = np.array(xcoords).transpose()
        ycoords = np.array(ycoords).transpose()

        # scale coords to lie in [0,1]^2
        xcoords, ycoords = self._scale_coords(xcoords, ycoords)

        # write out videoswarm files
        self._output_VS_files(table, header, xcoords, ycoords, 
            output_dir, csv_out, csv_no_index, csv_index_header, csv_headers)

    # scale coordinates for VideoSwarm interface
    def _scale_coords (self, xcoords, ycoords):

        # get range for x
        min_x = np.amin(xcoords)
        max_x = np.amax(xcoords)

        # get range for y
        min_y = np.amin(ycoords)
        max_y = np.amax(ycoords)

        # scale coordinates to be in [0,1]^2
        # if constant assign value of 1/2
        xcoords = (xcoords - min_x)
        if max_x > min_x:
            xcoords = xcoords / (max_x - min_x)
        else:
            xcoords = xcoords + 0.5

        ycoords = (ycoords - min_y)
        if max_y > min_y:
            ycoords = ycoords / (max_y - min_y)
        else:
            ycoords = ycoords + 0.5

        return xcoords, ycoords

    # write out VideoSwarm files
    def _output_VS_files(self, meta_data, expand_header, xcoords, ycoords, 
        output_dir, csv_out, csv_no_index, csv_index_header, csv_headers):

        # remove expanded column, if requested
        if self.args.remove_expand_col:
            meta_data.table = meta_data.table.drop(expand_header, axis=1)

        # write out movies.csv file
        movies_out = os.path.join(output_dir, csv_out)
        meta_data.to_csv(movies_out, index=csv_no_index, 
            index_label=csv_index_header, cols=csv_headers)

        # use samae root as csv_out
        csv_root = os.path.splitext(os.path.basename(csv_out))[0]

        # write out movies.xcoords file (use only float precision)
        xcoords_file_name = Path(os.path.join(output_dir, csv_root + '.xcoords')).as_posix()
        with open(xcoords_file_name, 'w', newline='', encoding='utf-8') as xcoords_file:
            csv_xcoords_file = csv.writer(xcoords_file)
            for i in xcoords.tolist():
                csv_xcoords_file.writerow(['{:f}'.format(x) for x in i])
        self.log.info("Saved file " + xcoords_file_name + ".")

        # write out movies.xcoords file (use only float precision)
        ycoords_file_name = Path(os.path.join(output_dir, csv_root + '.ycoords')).as_posix()
        with open(ycoords_file_name, 'w', newline='', encoding='utf-8') as ycoords_file:
            csv_ycoords_file = csv.writer(ycoords_file)
            for i in ycoords.tolist():
                csv_ycoords_file.writerow(['{:f}'.format(y) for y in i])
        self.log.info("Saved file " + ycoords_file_name + ".")

        # add time to first row of xcoords to make trajectories
        num_frames, num_movies = xcoords.shape
        vid_duration = num_frames / float(self.args.video_fps)
        time_row = np.linspace(0, vid_duration, num=num_frames)
        traj = np.ones((num_movies + 1, num_frames))
        traj[0, :] = time_row
        traj[1:, :] = xcoords.transpose()

        # write out movies.trajectories
        traj_file_name = Path(os.path.join(output_dir, csv_root + '.trajectories')).as_posix()
        with open(traj_file_name, 'w', newline='', encoding='utf-8') as traj_file:
            csv_traj_file = csv.writer(traj_file)
            for i in traj.tolist():
                csv_traj_file.writerow(['{:f}'.format(t) for t in i])

        self.log.info("Saved file " + traj_file_name + ".")

# if called from the command line display plugin specific command line arguments
if __name__ == "__main__":

    videoswarm = Plugin()
    videoswarm.parser.parse_args()


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This file provides a command line interface for performing
# dimension reduction of numerical simulation data.

# The interface for reduce is the same interface used for
# convert, but instead of converting files, a dimension
# reduction is performed.

# S. Martin
# 1/29/2021


# standard library imports

# command line arguments
import argparse

# logging and error handling
import logging
import sys

# output
import os
from pathlib import Path

# restart file
import pickle

# 3rd party imports
import numpy as np

# local imports
import slypi.ensemble as ensemble
import slypi.ensemble.algorithms.reduction as algorithms
from slypi.ensemble.utilities import Table as EnsembleTable

# set up argument parser
def init_parser():

    # define our own version of the slypi parser
    description = "Performs dimemsion reduction on ensemble data.  Uses Python-like %d[::] " + \
                  "notation, where %d[::] specifies a range of numbers in a file name. For " + \
                  'example "time_step_%d[0:10:2].vtp" would specify every other file from ' + \
                  '"time_step_0.vtp" to "time_step_9.vtp".  If individual time steps are ' + \
                  'provided as input, the results are combined into a single matrix and ' + \
                  'output.  The output file extension is .rd.npy.'
                  
    parser = ensemble.ArgumentParser (description=description)

    # get files to convert from command line
    parser.add_argument("--ensemble", help="Directory or directories to include in ensemble, "
                        "specified using the Python like %%d[::] notation described above.")
    parser.add_argument("--input-files", help="Files per ensemble directory to use as input "
                        "for reduction, specified using %%d[::] notation.  Note that these "
                        "files will be pre-fixed by the ensemble directories.")

    # or get files to convert from .csv file
    parser.add_argument("--csv-file", help="CSV file which specifies ensemble directories "
                        "and input files (alternate to using --ensemble and --input-files).")
    parser.add_argument("--csv-col", help="Column in CSV file where input files are "
                        "specified, can be either a string or an integer (1-based).")
    
    # input type (optional)
    parser.add_argument("--input-format", help="Format for input files (optional, inferred "
                        "from file extension if not provided).")

    # input model from file
    parser.add_argument("--input-model", help="Input dimension reduction model from .pkl file "
                        "(do not train a new model).")

    # output arguments and options
    parser.add_argument("--output-dir", help="Directory to place output.  All files will be "
                        "stored using directories that mirror those specified by --ensemble.")
    parser.add_argument("--output-file", help="File name for reduced data, the same name is "
                        "used for each simulation.")
    parser.add_argument("--output-model", help="Output dimension reduction model to provided "
                        "file (in output directory).")
    parser.add_argument("--over-write", action="store_true", help="Over-write output "
                        "directory if already present.")

    # output a csv file
    parser.add_argument("--csv-out", help="File name of output .csv file with file links for "
                        "reduced files (optional).  Will be written to output directory.")
    parser.add_argument("--csv-header", help="Name of output files header, needed only "
                        "if writing out a .csv file.")

    # output xy coordinate to csv file
    parser.add_argument("--xy-out", help="File name of output .csv file with the (x,y) " 
                        "coordinates (optional).  Will be written to output directory.")
    parser.add_argument("--xy-header", help="Root name of header for (x,y) coordinates " 
                        "columns in .csv file.")
    parser.add_argument("--xy-ps-tag", action="store_true", help="Add [XY Pair] tag to (x,y) "
                        "header output for parameter space models.")

    # batch option
    parser.add_argument("--file-batch-size", type=int, help="Train reduction model "
                        "incrementally using batches of files.  Not available for "
                        "all algorithms, see slypi.ensemble.algorithms.reduction --help for options.")

    # parallel option using ipyparallel
    parser.add_argument('--parallel', default=False, action="store_true", 
                        help="Use ipyparallel (must be available and running).")
    
    # restart file for parallel crashes
    parser.add_argument('--restart', help="File name to save intermediate results and then restart "
                        "from a crash (must also specify --output-model).")

    return parser

# check arguments
def check_arguments(log, args, algorithm=None):

    # check for command line input arguments
    if args.csv_file is None:

        # make sure the ensemble argument is present
        if args.ensemble is None:
            log.error("Ensemble directories are required.  Please use --ensemble and try again.")
            sys.exit(1)
        
        # make sure the input argument is present
        if args.input_files is None:
            log.error("Input files are required.  Please use --input-files and try again.")
            sys.exit(1)

        # disable args.csv_col (if provided)
        args.csv_col = None

    # otherwise check that ensemble arguments are not present
    else:

        if args.ensemble is not None or args.input_files is not None:
            log.error("Please specify either --csv-file or --ensemble, but not both.")
            sys.exit(1)

        # also check that csv column is provided
        if args.csv_col is None:
            log.error("Please specify --csv-col to provide input files.")
            sys.exit(1)

    # make sure input model file has .pkl extension
    if args.input_model is not None:
        if not args.input_model.endswith(".pkl"):
            log.error("Input model file must have .pkl extension.")
            sys.exit(1)

    # make sure the output directory is present
    if args.output_dir is None:
        log.error("Output directory must be specified.  Please use --output-dir and try again.")
        sys.exit(1)
    
    # make sure output file is specified
    if args.output_file is None:
        log.error("Output file name must be provided.  Please use --output-file and try again.")
        sys.exit(1)

    # check extension, if provided
    elif not args.output_file.endswith('.rd.npy'):
        log.error('Output file name must have extension ".rd.npy".')
        sys.exit(1)

    # make sure output model file has .pkl extension
    if args.output_model is not None:
        if not args.output_model.endswith(".pkl"):
            log.error("Output model file must have .pkl extension.")
            sys.exit(1)

    # if csv is to be written out, must also have column header
    if args.csv_out is not None:
        if args.csv_header is None:
            log.error("CSV header is required to output .csv file.  " +
                      "Please use --csv-header and try again.")
            sys.exit(1)

    # batch calculation checks
    if args.file_batch_size is not None:

        # check if batch size is non-negative
        if args.file_batch_size <= 0:
            log.error("Batch size must be an integer >= 1.")
            sys.exit(1)

        # check if algorithm is incremental
        if algorithm is not None:
            if not algorithm.is_incremental():
                if algorithm.time_align_dim() is None:
                    log.error("Dimension reduction algorithm selected is not incremental " +
                              "and can't be used in batch mode.  Please select a different " +
                              "algorithm and try again.")
                    sys.exit(1)

    # is restart file a .pkl file
    if args.restart is not None:

        # restart file must be .pkl file
        if not args.restart.endswith(".pkl"):
            log.error("Restart file must have .pkl extension.  Please use a different file " +
                    "name and try again.")
            sys.exit(1)

        # must also specify output model
        if args.output_model is None:
            log.error("Output model file (using --output-model) must be specified in addition to restart file.")
            sys.exit(1)

    # are there unknown arguments?
    if algorithm is not None:
        if len(algorithm.unknown_args()) > 0:
            log.error("Unrecognized arguments: %s.  Please try again." % 
                str(algorithm.unknown_args()))
            sys.exit(1)

# get simulation file list
def get_sim_files (log, args, ensemble_dirs, mirror_dirs, ensemble_table):

    # get simulations files in parallel if requested
    sim_files = ensemble_table.ensemble_files(ensemble_dirs, args.parallel) 
    
    # count number of files per ensemble
    file_counts = []
    for i in range(len(ensemble_dirs)):

        # count files
        file_counts.append(len(sim_files[i]))

        # report on files found
        log.info("Found %d file(s) in ensemble directory %s." % 
            (len(sim_files[i]), ensemble_dirs[i]))

    # construct output file names/check sim files
    output_files = []
    for i in range(len(ensemble_dirs)):

        # print error if no files found for conversion
        if len(sim_files[i]) == 0:
            log.error("No files to reduce, please provide existing files for input.")
            sys.exit(1)

        # output file name for each ensemble
        output_files.append(Path(os.path.join(mirror_dirs[i], args.output_file)).as_posix())

    # check that time counts are uniform
    if min(file_counts) < max(file_counts):
        log.error("Simulations have different number of time step files.  Dimension " + \
                  "reduction not performed, no files written.")
        sys.exit(1)

    return sim_files, output_files

# get files for a single time step
def get_time_files (sim_files, time_step, file_batch_size):

    # sim_files is organized as a [sim][time]
    # go through all simulations and pick out desired time step files
    time_files = []
    num_sim = len(sim_files)
    for sim in range(num_sim):
        time_files.append(sim_files[sim][time_step])

    # split time files into batches
    return split_batch_files(time_files, file_batch_size)

# split a list of files into batches
def split_batch_files (batch_files, file_batch_size):

    # split into array of batch sizes
    batch_inds = list(range(0, len(batch_files), file_batch_size))
    if batch_inds[-1] < len(batch_files):
        batch_inds += [len(batch_files)]
    
    # get files by batch
    batches = []
    for i in range(len(batch_inds)-1):
        batches.append(batch_files[batch_inds[i]:batch_inds[i+1]])
    
    return batches

# get a batch of file names
def get_batch_files (sim_files, file_batch_size):

    # concatenate all sim files, in order
    batch_files = [file for sim in sim_files for file in sim]

    # return all files if no batches
    if file_batch_size is None:
        return [batch_files]
    
    # otherwise split into array of batch sizes
    else:
        return split_batch_files(batch_files, file_batch_size)

# read in a batch of files and pre-process, 
# return number of time points per file
def get_batch (log, plugin, batch_files, file_type, 
    parallel, flatten):

    # get batch of data
    data = plugin.read_file_batch(batch_files, file_type=file_type, 
        parallel=parallel, flatten=flatten)
    
    # pre-process and count time points
    time_counts = []
    for i in range(len(batch_files)):

        # make sure data is numpy array
        data[i] = np.asarray(data[i])
        
        # keep track of time points per file
        if len(data[i].shape) == 1:
            time_counts.append(1)
        else:
            time_counts.append(data[i].shape[0])

    # check that time points per file are identical
    num_time = max(time_counts)
    if min(time_counts) < num_time:
        log.error("Simulation time steps per file are different.  Dimension " + \
                    "reduction not performed, no files written.")
        sys.exit(1)

    # convert to 2D matrix for dimension reduction
    data_to_reduce = np.asarray(data)
    if num_time > 1:
        data_to_reduce = np.vstack(data)

    return data_to_reduce, num_time

# compute time indices into array of stacked simulation data
def compute_time_inds(num_time_steps, num_sim, time_step):

    # gather simulation data per time step
    time_inds = []
    for j in range(num_sim):
        time_inds.append(j * num_time_steps + time_step)

    return time_inds

# train a single time step, using algorithm.fit
def train_time_step(log, args, algorithm, data, time_step):
    
    # train dimension reduction model, if not loaded
    if args.input_model is None:
        log.info("Training dimension reduction model for time step %d." % time_step)
        algorithm.fit(data, time_step=time_step)

    # project data
    log.info("Projecting data to lower dimension for time step %d." % time_step)
    time_reduced_data = algorithm.transform(data, time_step=time_step)

    return time_reduced_data

# project data in batches
def project_data_batches(log, args, plugin, algorithm, batch_files, 
    time_step=0, flatten=True):

    # project data in batches
    num_batches = len(batch_files)
    reduced_data = []
    for i in range(num_batches):

        # read batch files into matrix
        data_to_reduce, num_time = get_batch(log, plugin,
            batch_files[i], args.input_format, args.parallel, flatten)

        # collect reduced data
        log.info("Projecting data to lower dimension.")
        reduced_data.append(algorithm.transform(data_to_reduce, time_step=time_step))
    
    # make each batch into one matrix
    reduced_data = np.vstack(reduced_data)

    return reduced_data, num_time

# convert from time aligned list of matrices to 
# all simulations concatenated matrix
def convert_time_sim(time_aligned_data, num_time_steps, num_sim, num_dim):

    reduced_data = np.zeros((num_time_steps * num_sim, num_dim))
    for i in range(num_time_steps):

        # gather simulation data per time step
        time_inds = compute_time_inds(num_time_steps, num_sim, i)

        # distrubte aligned data into matrix
        reduced_data[time_inds,:] = time_aligned_data[i]

    return reduced_data

# performs dimension reduction on ensemble
def reduce(arg_list=None):

    # initialize parser
    parser = init_parser()

    # parse arguments
    if arg_list is not None:

        # parse in case called with arg_list
        args, arg_list = parser.parse_args(arg_list)

    else:

        # command line parser
        args, arg_list = parser.parse_args()

    # start logger
    ensemble.init_logger(log_file=args.log_file, log_level=args.log_level)
    log = logging.getLogger("slypi.ensemble.reduce")
    log.debug("Started reduce.")

    # check arguments
    check_arguments(log, args)

    # import and initialize plugin
    plugin, arg_list = ensemble.plugin(args.plugin, arg_list)

    # initialize dimension reduction algorithm 
    algorithm = algorithms.DimensionReduction(model_file=args.input_model, 
                                              arg_list=arg_list)

    # re-check arguments with algorithm initialized
    check_arguments(log, args, algorithm=algorithm)
    
    # create ensemble table
    ensemble_table = EnsembleTable(log, csv_file=args.csv_file,
        ensemble_spec=args.ensemble, file_spec=args.input_files, 
        header="reduce")

    # get ensemble directories/file specs
    if args.csv_file is not None:
        ensemble_dirs = ensemble_table.get_col(args.csv_col)
    else:
        ensemble_dirs = ensemble_table.get_col("reduce")

    # quit if no directories found
    num_ensemble_dirs = len(ensemble_dirs)
    if num_ensemble_dirs == 0:
        log.error("No ensemble directories found.  " +
                  "Please identify existing directories and try again.")
        sys.exit(1)
    else:
        log.info("Found %d ensemble directory(ies)." % num_ensemble_dirs)

    # mirror directories to output direoctory
    mirror_dirs = ensemble_table.mirror_directories(
        args.output_dir, ensemble_dirs, args.over_write)
    
    # if directory already exists, quit
    if mirror_dirs is None:
        log.error("Output directory already exists, use --over-write if " +
            "you want to over-write or add to previous contents.")
        sys.exit(1)
    log.info("Output directory created or already exists with mirrored file structure.")

    # check for restart file option
    restart_file_loaded = False
    if args.restart is not None:

        # check for existence of restart file
        restart_file = os.path.join(args.output_dir, args.restart)
        if os.path.exists(restart_file):

            # get sim files and start index from restart file
            with open(restart_file, 'rb') as handle:
                restart_time, restart_batch, sim_files, output_files, time_reduced_data = \
                    pickle.load(handle)
            log.info("Loaded restart data from %s." % restart_file)
            log.info("Restarting at time-aligned step %d, batch %d. " % (restart_time, restart_batch))

            # get model from output model
            model_file = os.path.join(args.output_dir, args.output_model)
            algorithm = algorithms.DimensionReduction(model_file=model_file)

            restart_file_loaded = True

    # if restart file was not loaded, recompute and start from 0
    if not restart_file_loaded:

        # get simulation and output files
        sim_files, output_files = get_sim_files (log, args, 
            ensemble_dirs, mirror_dirs, ensemble_table)

        # set restart to 0
        restart_time = 0
        restart_batch = 0

        # time-reduced data for time-aligned case
        time_reduced_data = []
    
    # count up simulations and time steps
    num_sim = len(sim_files)
    num_files = len(sim_files[0])

    # determine batches
    batch_files = get_batch_files (sim_files, args.file_batch_size)
    num_batches = len(batch_files)

    # check for rd.npy file type
    rd_npy_type = True
    for i in range(num_batches):
        for j in range(len(batch_files[i])):
            if not batch_files[i][j].endswith('.rd.npy'):
                rd_npy_type = False
    
    # if rd.npy file type, do not flatten data
    flatten = True
    if rd_npy_type:
        flatten = False

    # each reduction option has to compute reduced_data
    # ordered same as sim_files, but flattened, and num_time
    # number of time steps per file

    # do reduction differently for different cases
    if algorithm.time_align_dim() is None:
        if args.file_batch_size is None:
            
            # 1st case is bulk reduction in memory

            # read all files into matrix
            data_to_reduce, num_time = get_batch(log, plugin,
                batch_files[0], args.input_format, args.parallel, flatten)
            
            # train dimension reduction model, if not loaded
            if args.input_model is None:
                log.info("Training dimension reduction model.")
                algorithm.fit(data_to_reduce)

            # reduce data
            log.info("Projecting data to lower dimension.")
            reduced_data = algorithm.transform(data_to_reduce)

        else:

            # 2nd case is bulk in batches

            # train in batches (unless model has been loaded)
            if args.input_model is None:
                for i in range(restart_batch, num_batches):

                    # read batch files into matrix
                    data_to_reduce, num_time = get_batch(log, plugin,
                        batch_files[i], args.input_format, args.parallel, flatten)

                    # incremental model training
                    log.info("Training dimension reduction model, batch %d." % i)
                    algorithm.partial_fit(data_to_reduce)

                    # if restart file, save partial results
                    if args.restart is not None:

                        # save model file
                        model_file = os.path.join(args.output_dir, args.output_model)
                        algorithm.save(model_file)

                        # save restart file
                        with open(restart_file, 'wb') as handle:
                            pickle.dump([0, i+1, sim_files, output_files, []], handle)
                        log.info("Saved restart file: %s." % restart_file)

            # compute reduced data, also in batches
            reduced_data, num_time = project_data_batches(log, args, plugin, algorithm, batch_files,
                flatten=flatten)

    # time aligned cases
    else:

        if args.file_batch_size is None:
            
            # 3rd case is time aligned in memory
            
            # read all files into matrix
            data_to_reduce, num_time = get_batch(log, plugin,
                batch_files[0], args.input_format, args.parallel, flatten)

            # perform dimension reduction per time step
            num_time_steps = num_time * num_files
            for i in range(restart_time, num_time_steps):
                
                # gather simulation data per time step
                time_inds = compute_time_inds (num_time_steps, num_sim, i)

                # train dimension reduction model, if not loaded
                time_reduced_data.append(train_time_step(log, args, 
                    algorithm, data_to_reduce[time_inds,:], i))
                
                # if restart file, save partial results
                if args.restart is not None:

                    # save model file
                    model_file = os.path.join(args.output_dir, args.output_model)
                    algorithm.save(model_file)

                    # save restart file
                    with open(restart_file, 'wb') as handle:
                        pickle.dump([i, 0, sim_files, output_files, time_reduced_data], handle)
                    log.info("Saved restart file: %s." % restart_file)

            # time align
            log.info("Time aligning reduced data.")
            time_aligned_data = algorithm.time_align(time_reduced_data, 
                                    compute_rotations=(args.input_model is None))

            # put back into matrix form
            reduced_data = convert_time_sim(time_aligned_data, num_time_steps, 
                num_sim, algorithm.num_dim())

        else:

            # 4th case is time aligned in batches

            # proceed in batches per time step
            for i in range(restart_time, num_files):

                # get time step files 
                time_files = get_time_files(sim_files, i, args.file_batch_size)

                # check that algorithm is incremental if more than one batch
                num_batches = len(time_files)
                if num_batches > 1 and not algorithm.is_incremental():
                    log.error("There are too many simulation files per time step for the given " +
                              "file batch size.  Please use an inremental algorithm or a larger " +
                              "file batch size.")
                    sys.exit(1)

                # train reduction in batches
                for j in range(num_batches):

                    # load batch of data
                    data_to_reduce, num_time = get_batch(log, plugin,
                        time_files[j], args.input_format, args.parallel, flatten)
                    
                    # for a single batch, use algorithm.fit
                    if num_batches == 1:
                        time_reduced_data.append(train_time_step(log, args, 
                            algorithm, data_to_reduce, i))

                    # for multiple batches, use algorthm.partial_fit,
                    # if model is not already loaded
                    elif args.input_model is None:
                        log.info("Training dimension reduction model, batch " + str(j) +", " +
                                 "time step " + str(i) + ".")
                        algorithm.partial_fit(data_to_reduce, time_step=i)

                # project data, also in batches
                if num_batches > 1:
                    proj_data, _ = project_data_batches(log, args, 
                        plugin, algorithm, time_files, time_step=i, flatten=flatten)
                    time_reduced_data.append(proj_data)
                
                # if restart file, save partial results
                if args.restart is not None:

                    # save model file
                    model_file = os.path.join(args.output_dir, args.output_model)
                    algorithm.save(model_file)

                    # save restart file
                    with open(restart_file, 'wb') as handle:
                        pickle.dump([i, 0, sim_files, output_files, time_reduced_data], handle)
                    log.info("Saved restart file: %s." % restart_file)

            # time align
            log.info("Time aligning reduced data.")
            time_aligned_data = algorithm.time_align(time_reduced_data, 
                                    compute_rotations=(args.input_model is None))

            # put back into matrix form
            num_time_steps = num_time * num_files
            reduced_data = convert_time_sim(time_aligned_data, num_time_steps, 
                num_sim, algorithm.num_dim())
    
    # save data
    for i in range(num_sim):

        # each file is a matrix (or vector)
        reduced_data_j = []
        for j in range(num_files):

            # get index range in data per file
            start_index = (i * num_files + j * num_time) * num_time
            end_index = start_index + num_time

            # collectreduced data representation per file
            reduced_data_j.append(reduced_data[start_index:end_index,:])
        
        # write out reduced data per simulation        
        plugin.write_file(np.vstack(reduced_data_j), output_files[i])
        log.info("Saved file %s." % output_files[i])
        
    # save links
    if args.csv_out is not None:

        # add new column to ensemble table
        ensemble_table.add_col(output_files, args.csv_header)

        # put .csv file in output directory
        ensemble_table.to_csv(args.csv_out, output_dir=args.output_dir, 
            cols=[args.csv_header])

    # save xy coords, if requested (and possible)
    if args.xy_out is not None:

        # check for one time step per simulation
        if num_time * num_files > 1:
            log.warning("More than one time step per simulation, could not output (x,y) "
                        "coordinates.")
        
        # check for both x and y coordinates
        elif reduced_data.shape[1] < 2:
            log.warning("Less than two coordinates in reduction, could not output (x,y) " 
                        "coordinates.")
        
        # output coordinates
        else:

            # get (x,y) coords
            x_coords = np.squeeze(np.asarray(reduced_data))[:,0]
            y_coords = np.squeeze(np.asarray(reduced_data))[:,1]

            # construct x, y headers
            x_header = "X" if args.xy_header is None else args.xy_header + " X"
            y_header = "Y" if args.xy_header is None else args.xy_header + " Y"

            # pre-pend ps tags, if requested
            if args.xy_ps_tag:
                x_header = "[XYpair X]" + x_header
                y_header = "[XYpair Y]" + y_header

            # add new columns to ensemble table
            ensemble_table.add_col(x_coords, x_header)
            ensemble_table.add_col(y_coords, y_header)

            # put .csv file in output directory
            ensemble_table.to_csv(args.xy_out, output_dir=args.output_dir,
                cols=[x_header, y_header])

    # save model
    if args.output_model is not None:
        algorithm.save(os.path.join(args.output_dir, args.output_model))
    

# entry point for command line call
if __name__ == "__main__":
    reduce()


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# Including __init__.py to make plugins accessible as packages.

# S. Martin
# 12/17/2020


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This module contains code to convert a file formats, in particular to output movies
# in mp4 Slycat readable format.

# standard library imports
import os
from pathlib import Path

# for writing mp4
import imageio

# for exporting images
from PIL import Image
from matplotlib import cm

# 3rd party imports
import numpy as np
import pandas as pd

# materials knowledge system
# for exporting images
from matplotlib import cm

# local imports
import slypi

# defaults for user arguments
JPG_QUALITY = 95
VIDEO_FPS = 25

# functions for jpg convert specific operations
# class must be named Plugin
class Plugin(slypi.ensemble.PluginTemplate):

    # initialize jpg convert command line argument parser
    def __init__(self):

        # describe jpg convert plugin
        description = "The convert plugin provides support for the converting between " + \
                      "different file formats."

        # set up parser
        super().__init__(description=description)

    # add any extra command line arguments
    def add_args(self):

        self.parser.add_argument("--suffix", default="", 
                                 help="Suffix to describe output file(s), "
                                      'e.g. "--suffix phase_field".  The suffix name is included in the name '
                                      "of any output file.")

        self.parser.add_argument("--binary", action="store_true", help="Converts field "
                                                                       "variable to binary by clipping anything less than 0 to 0 and anyting "
                                                                       "greater than 0 to 1.")

        self.parser.add_argument("--color-scale", nargs=2, type=float, help="Gives the color "
                                                                            "scale for the field variable when creating jpg or mp4, e.g. "
                                                                            '"--color-scale 0 1" for min color value of 0 and max color value of 1. '
                                                                            "Note that values above and below the color scale are automatically clipped.")

        self.parser.add_argument("--output-format", help="The output format options "
                                                         'recognized by the convert plugin include: "npy" -- saves the '
                                                         'field variable for a single timestep to a numpy array; "sim.npy" -- '
                                                         'saves the field variable for every timestep in a simulation to a 3D '
                                                         'numpy array; "rd.npy" -- saves the reduced dimensional representation '
                                                         'to a numpy array (can be either a time step matrix or a 3D full simulation '
                                                         'matrix); "jpg" -- saves a .jpg image of the field variable for a single '
                                                         'timestep; "mp4" -- saves a .mp4 movie of the field variable for every timestep '
                                                         'in a simulation.')

        self.parser.add_argument("--output-quality", type=int, default=JPG_QUALITY,
                                 help="Quality of jpg image, as a number between 1 and 95 (only relevant "
                                      "if outputing images, defaults to %s)." % str(JPG_QUALITY))

        self.parser.add_argument("--video-fps", type=float, default=VIDEO_FPS,
                                 help="Number of frames per second for video creation, defaults to %s)." %
                                      str(VIDEO_FPS))

        self.parser.add_argument("--write-raw-video", action="store_true", help="Create a video using the provided images unprocessed.")

    # check arguments, error if incorrect input
    def check_args(self, args):

        # check .jpg output_quality
        if args.output_quality < 1 or args.output_quality > 95:
            self.log.error("Quality option --output-quality must be between 1 and 95.  " +
                           "Please try again.")
            raise ValueError("output quality must be between 1 and 95.")

        # check fps is > 0
        if args.video_fps <= 0:
            self.log.error("Video frames per second --video-fps must be > 0.")
            raise ValueError("video fps must be > 0.")

    # initialize any local variables from command line arguments
    def init(self, args):

        # save args for later use
        self.args = args

    # make binary version of matrix
    def _binary(self, data):

        data[data < 0] = 0
        data[data > 0] = 1

        return data

    # clip matrix values and scale matrix to [0,1] 
    def _scale_matrix(self, point_data):
        
        # default to scaling to min/max of data
        point_data_min = point_data.min()
        point_data_max = point_data.max()
        
        # color-scale is provided clip data and use color scale for min/max
        if self.args.color_scale is not None:

            # clip data to min and max of color scale
            point_data[point_data < self.args.color_scale[0]] = self.args.color_scale[0]
            point_data[point_data > self.args.color_scale[1]] = self.args.color_scale[1]

            # use color scale for min and max to scale
            point_data_min = self.args.color_scale[0]
            point_data_max = self.args.color_scale[1]

        # scale data to [0,1] for image creation
        point_data = point_data - point_data_min    
        if point_data_min < point_data_max:
            point_data = point_data/(point_data_max - point_data_min)
        
        return point_data, point_data_min, point_data_max

    # convert matrix to mp4 frame
    def _matrix_to_mp4(self, data):

        # scale matrix to [0,1]
        point_data_out, _, _ = self._scale_matrix(data)

        # convert to image and remove alpha
        return np.delete(np.uint8(cm.jet(point_data_out) * 255), 3, 2)

    # read npy and sim.npy (also npy) formats
    def read_file(self, file_in, file_type=None):

        # check file extension, if not provided
        if file_type is None:

            # npy file type
            if file_in.endswith('.npy'):
                file_type = 'npy'
            
            if file_in.endswith('.npz'):
                file_type = 'npz'

            if file_in.endswith('.png'):
                file_type = 'png'

        # check if we have npy or sim.npy
        if file_type == 'npy':
            
            # read npy file
            try:
                data = np.load(file_in)
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("Could not read " + file_in + " as a .npy file.")

        # is it .npz?
        elif file_type == 'npz':

            # read npz file
            try:
                data = np.load(file_in)
                data = data['arr_0']
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("Could not read " + file_in + " as a .npy file.")

        # is it an image file?
        elif file_type == 'png':

            try:
                img = Image.open(file_in)
                data = np.array(img)
            except ValueError:
                self.log.error("Could not read " + file_in + "as a .png file.")
                raise ValueError("Could not read " + file_in + " as a .png file.")
            
        # otherwise default to mesh
        else:
            data = super().read_file(file_in, file_type)

        return data

    # write out .npy format or .jpg
    def write_file(self, data, file_out, file_type=None):

        # check for npy output format with color scale
        if file_type=='npy':
            if self.args.color_scale is not None:
                self.log.error("Can't use --color-scale with .npy output.")
                raise ValueError("can't use --color-scale with .npy output.")

        # infer file type, if not provided
        if file_type is None:

            if file_out.endswith('.npy'):
                file_type = 'npy'

            if file_out.endswith('.npz'):
                file_type = 'npz'

            if file_out.endswith('.jpg'):
                file_type = 'jpg'

        # check for npy or jpg file output (data is matrix format)
        if file_type in ['npy', 'npz', 'jpg']:

            # convert to binary, if requested
            if self.args.binary:
                data = self._binary(data)

            # write out field variable as npy, npz or jpg
            if file_type == "npy":

                # save as numpy
                np.save(file_out, data)
            
            # write out field variable as npz
            elif file_type == "npz":

                # save as compressed numpy
                np.savez_compressed(file_out, np.asarray(data, dtype=np.float16))

            # for jpg we need to scale data and use a colormap
            elif file_type == "jpg":

                # is it already RGB?
                if len(data.shape) == 3:
                    img = Image.fromarray(data.astype(np.uint8), 'RGB')
                    img.save(file_out, quality=self.args.output_quality)
                
                else:
                    # convert to [0,1]
                    data, _, _ = self._scale_matrix(data)
                    
                    # convert to standard jet color map image
                    img = Image.fromarray(np.uint8(cm.jet(data) * 255))

                    # save image
                    img.convert("RGB").save(file_out, quality=self.args.output_quality)

        else:

            # default to meshio output
            super().write_file(data, file_out, file_type=file_type)

    # over-riding convert_files to generate sim.npy and mp4 files
    def convert_files(self, file_list, output_dir, output_type, input_type=None):

        # check for dakota tabular.dat file
        if len(file_list) == 1:
            if file_list[0].endswith(".dat") or input_type == "dat":
                if output_type == "csv":

                    # read tabular.dat file
                    df = pd.read_csv(file_list[0], sep=r'\s+')

                    # pre-pend workdir.x as identification column
                    eval_id = df['%eval_id']
                    file_path = os.path.split(file_list[0])[0]
                    workdirs = [os.path.join(file_path, 'workdir.' + str(id))
                                for id in eval_id]
                    df.insert(0, column='', value=workdirs)
                    
                    # save as tabular.csv file
                    root = os.path.split(file_list[0])[1][:-4]
                    output_file = Path(os.path.join(output_dir, root + '.csv')).as_posix()
                    df.to_csv(output_file, index=False)

                    return [output_file]

        # check for sim.npy or mp4
        if output_type == "sim.npy" or output_type == "mp4":

            # go through file list and load data
            file_data = []
            num_files = len(file_list)
            for file_to_add in file_list:

                point_data_out = imageio.imread(file_to_add)

                # convert to binary, if requested
                if self.args.binary:
                    point_data_out = self._binary(point_data_out)

                # make sure we only have at most one 3D matrix
                matrix_dim = len(point_data_out.shape)

                # add as matrix
                if output_type == "sim.npy":

                    # check that it's 2D matrix:
                    if matrix_dim == 2:
                        file_data.append(point_data_out)

                    # otherwise keep data the same
                    else:
                        file_data = point_data_out

                # add as jet colormap image
                else:

                    # convert each frame for 2D
                    if matrix_dim == 2:

                        # convert to mp4 frame
                        file_data.append(self._matrix_to_mp4(point_data_out))

                    # check if frame is RGB
                    elif point_data_out.shape[2] == 3:

                        # keep data the same
                        file_data.append(point_data_out)
                    
                    else:

                        # convert all frames for 3D
                        for i in range(point_data_out.shape[0]):
                            file_data.append(self._matrix_to_mp4(point_data_out[i, :, :]))

            # create file name by combining root file names with field variable
            if num_files > 1:

                # get common file prefix
                common_prefix = os.path.basename(os.path.commonprefix(file_list))

                # remove trailing zeros, if they exist
                common_prefix = common_prefix.rstrip('0')

                # join to output directory and add field variable
                file_out = Path(os.path.join(output_dir, common_prefix +
                                self.args.suffix + "." + output_type)).as_posix()

            # if one file, change extension to output type
            else:

                file_name = os.path.basename(file_list[0])

                # check for .sim.npy file (very likely)
                if file_name.endswith('.sim.npy'):
                    file_root, file_ext = file_name.split('.sim.npy')

                # otherwise split normally
                else:
                    file_root, file_ext = os.path.splitext(file_name)

                file_out = Path(os.path.join(output_dir, file_root + "." + 
                                             output_type)).as_posix()

            # write out as sim.npy
            if output_type == "sim.npy":
                np.save(file_out, np.asarray(file_data))

            # write out as mp4
            else:

                # open a video writer with web browser codecs
                writer = imageio.get_writer(file_out, format="FFMPEG", mode="I", 
                    fps=self.args.video_fps, output_params=['-force_key_frames', 
                    '0.0,0.04,0.08']) #, '-vcodec', 'libx264', '-acodec', 'aac'])

                # write frames to movie
                if not self.args.write_raw_video:
                    for i in range(len(file_data)):
                        writer.append_data(file_data[i])
                else:
                    # add raw files to imageio video writer
                    for f in file_list:
                        writer.append_data(imageio.imread(f))

                writer.close()

            # must return a list of files written
            return [file_out]

        else:
        
            # revert to 1-1 file conversion
            return super().convert_files(file_list, output_dir, output_type, input_type=input_type)


# if called from the command line display plugin specific command line arguments
if __name__ == "__main__":
    plugin = Plugin()
    plugin.parser.parse_args()



# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This is a plugin that provides support for converting dimension
# reduction results to Parameter Space format.  In particular
# it is to be called using the --expand option from table.py.

# S. Martin
# 4/21/2021


# standard library imports
import argparse
import csv
import os

# 3rd party imports
import numpy as np
import pandas as pd
from PIL import Image

# auto-correlation
try:
    import pymks
except ImportError:
    pymks = None

# local imports
from slypi.ensemble import utilities
import slypi


# functions for specific operations
# class must be named Plugin
class Plugin(slypi.ensemble.PluginTemplate):

    # initialize command line argument parser
    def __init__(self):

        # describe plugin
        description = "The parameter space plugin provides the ability to convert " + \
            "dimension reduction results to Slycat Parameter Space " + \
            "format."

        # set up parser
        super().__init__(description=description)

    # add any extra command line arguments
    def add_args(self):

        self.parser.add_argument("--num-coords", type=int, default=2, help="Number of coordinates " +
            "from input files to include in .csv output file.")

        self.parser.add_argument("--remove-expand-col", action="store_true", default=False,
            help="Remove the expanded column when writing out parameter space file.")

        self.parser.add_argument("--include-original-index", action="store_true", default=False,
            help="Add original (repeated) index to expanded output .csv file.")

        self.parser.add_argument("--binary", action="store_true", help="Converts field "
            "variable to binary by clipping anything less than 0 to 0 and anyting "
            "greater than 0 to 1.")

        self.parser.add_argument("--auto-correlate", action="store_true", help="Performs "
            "auto-correlation as a pre-processing for dimension reduction (note this option "
            "requires the --binary flag to be used).")

        self.parser.add_argument("--scale", action="store_true", help="Scale values to [-0.5,0.5].")

    # check arguments, error if incorrect input
    def check_args(self, args):

        if args.num_coords is not None:
            if args.num_coords < 2:
                self.log.error("Number of coordinates to output must be >= 2.")
                raise ValueError("number of coordinates to output must be >= 2 (see " +
                                 "slypi.ensemble.plugins.parameter_space --help).")

        # check that binary is active if auto-correlate is selected
        if args.auto_correlate:

            # check if pymks is available
            if not pymks:
                self.log.error("The pymks module has not been installed, " + 
                                 "cannot perform auto-correlation. " +
                                 "Use pip install slypi[auto].")
                raise ValueError("The pymks module has not been installed, " + 
                                 "cannot perform auto-correlation. " +
                                 "Use pip install slypi[auto].")

            # check that binary option is enabled
            if not args.binary:
                self.log.error("Auto-correlation requires binary input, please use --binary " +
                               "and try again.")
                raise ValueError("auto-correlation requires binary flag (see " +
                                 "slypi.ensemble.plugins.vs --help.")
            

    # initialize any local variables from command line arguments
    def init(self, args):

        # save args for later use
        self.args = args

    # convert from 3 channel image to numpy 
    def _convert_img (self, data):

        if len(data.shape) == 3:

            # convert to black and white image then numpy array
            img = Image.fromarray(data).convert('L')
            data = np.array(img)

        return data
    
    # scale values between [0,1]
    def _scale(self, data):

        min = data.min()
        max = data.max()

        data = (data - min)/(max - min) - 0.5

        return data

    # make binary version of matrix
    def _binary(self, data):

        data[data < 0] = 0
        data[data > 0] = 1

        return data

    # perform pre-processing on jpg data (n,n,3)
    def preprocess (self, data, flatten=True):

        # if image, convert to matrix using jet map
        data = self._convert_img(data)

        # scale data, if requested
        if self.args.scale:
            data = self._scale(data)

        # convert to binary, if requested
        if self.args.binary:
            data = self._binary(data)

        # do auto-correlation, if requested
        if self.args.auto_correlate:
            
            # set up hat basis for pymks
            p_basis = pymks.bases.PrimitiveBasis(n_states=2)

            # check if this data is 2D
            if len(data.shape) == 2:

                # keep rows and columns
                num_rows, num_cols = data.shape

                # add extra dimension for 2D
                data = np.expand_dims(data, axis=0)
            
            # do auto-correlation with pymks (use binary data)
            space_stats = pymks.stats.correlate(data, p_basis, 
                                                periodic_axes=(0, 1), 
                                                correlations=[(0, 0)])

            # re-shape each time step into vector
            space_stats_vec = np.reshape(space_stats,[space_stats.shape[0],
                                         space_stats.shape[1] * space_stats.shape[2],
                                         space_stats.shape[-1]])
            data_vec = np.squeeze(space_stats_vec)

            # re-shape to 2D if we don't want flattened data
            if not flatten:

                data_vec = np.reshape(data_vec, (num_rows, num_cols))
        
        else:

            # default behaviour is to reshape matrix into one simulation per row
            if flatten:

                if len(data.shape) == 2:
                    data_vec = np.reshape(data, [data.shape[0] * data.shape[1]])

                else:
                    data_vec = np.reshape(data, [data.shape[0], 
                                        data.shape[1] * data.shape[2]])

            # otherwise, keep as matrix
            else:
                data_vec = data
                
        # make sure there are no singleton dimensions
        data_vec = np.squeeze(data_vec)
        
        return data_vec

    # over-ride file read, accept .jpg or .rd.npy files
    def read_file(self, file_in, file_type=None):

        # is it an npy file?
        if file_in.endswith('npy'):

            # read npy file
            try:
                data = np.load(file_in)
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("could not read " + file_in + " as a .npy file.")

        # is it .npz?
        elif file_in.endswith('.npz'):

            # read npz file
            try:
                data = np.load(file_in)
                data = data['arr_0']
            except ValueError:
                self.log.error("Could not read " + file_in + " as a .npy file.")
                raise ValueError("could not read " + file_in + " as a .npy file.")
            
        elif file_in.endswith('.png'):

            try:
                img = Image.open(file_in)
                data = np.array(img)
            except ValueError:
                self.log.error("Could not read " + file_in + "as a .png file.")
                raise ValueError("Could not read " + file_in + " as a .png file.")
            
        else:
            self.log.error("The parameter space plugin accepts only .png, .rd.npy or .npz files.")
            raise TypeError("parameter space plugin accepts only .png, .rd.npy or .npz files.")
        
        self.log.info("Read file " + file_in + ".")
        
        return data

    # write out .npy format
    def write_file(self, data, file_out, file_type=None):

        # infer file type, if not provided
        if file_type is None:
            if file_out.endswith('.npy'):
                file_type = 'npy'
            elif file_out.endswith('.npz'):
                file_type = 'npz'

        # check for npy file output (data is matrix format)
        if file_type == 'npy':
            np.save(file_out, data)
        
        elif file_type == 'npz':
            np.savez(file_out, data)

        else:

            # default to meshio output
            super().write_file(data, file_out, file_type=file_type)
    
    # over-riding expand to generate videoswarm input files
    def expand(self, table, header, file_list, **kwargs):
        """
        Outputs the Parameter Space input file from a set of .rd.npy files.

        Optional arguments include (see also table.py options):
            output_dir -- output directory
            csv_out -- file name for csv output
            csv_no_index -- do not include pandas index in movies.csv
            csv_index_header -- use this header name for index column
            csv_headers -- output only these columns to movies.csv

        Other arguments passed via kwargs are ignored.
        """
        
        # decode extra arguments
        output_dir = kwargs.get("output_dir")
        csv_out = kwargs.get("csv_out")
        csv_no_index = kwargs.get("csv_no_index", False)
        csv_index_header = kwargs.get("csv_index_header", None)
        csv_headers = kwargs.get("csv_headers", None)

        # put csv_out in output directory
        csv_out = os.path.join(output_dir, csv_out)

        # file_list is in order of table, expecting .rd.npy files
        reduced_coords = []
        for file_name in file_list:
            
            # read file
            data = self.read_file(file_name)

            # check that data is at least two dimensional
            if data.shape[1] < 2:
                self.log.error("Data in " + file_name + " contains less than " +
                               "two dimensions.")
                raise ValueError("reduced dimension data must have at least two dimensions.")
            
            # check that number of dimension is not too large
            if data.shape[1] < self.args.num_coords:
                self.log.warning("Data in " + file_name + " contains less than " +
                                 "requested number of dimensions to output -- defaulting " +
                                 "to " + str(data.shape[1]) + " dimensions.")
                self.args.num_coords = data.shape[1]

            # collate data (each .rd.npy file contains time steps 
            # per simulation as rows and reduced data as columns)
            reduced_coords.append(data[:,0:self.args.num_coords])

        # convert reduced coords into list per dimension
        reduced_coords_per_dim = [[] for i in range(self.args.num_coords)]
        for i in range(self.args.num_coords):
            for j in range(len(reduced_coords)):
                reduced_coords_per_dim[i].append(reduced_coords[j][:,i])

        # keep track of original index
        original_index = []
        for j in range(len(reduced_coords)):
            original_index.append((j+1) * np.ones(reduced_coords[j].shape[0]))

        # write out files
        self._output_PS_files(table, header, reduced_coords_per_dim,
            original_index, csv_out, csv_no_index, csv_index_header, csv_headers)

    # write out Parameter Space files
    def _output_PS_files(self, meta_data, expand_header, reduced_coords_per_dim, 
        original_index, csv_out, csv_no_index, csv_index_header, csv_headers):
        
        # remove expanded column, if requested
        if self.args.remove_expand_col:
            meta_data.table = meta_data.table.drop(expand_header, axis=1)

        # add a new column for each requested dimension
        for i in range(self.args.num_coords):

            # for 1st dimension, expand table
            if i == 0:

                # add either original index + first dimension column
                if self.args.include_original_index:
                    expanded_data = utilities.explode(self.log, meta_data,
                        'Original Index', original_index)
                    expanded_data.add_col(np.concatenate(reduced_coords_per_dim[i]).tolist(),
                        expand_header + " Dimension " + str(i+1))

                # or just first dimension column
                else:
                    expanded_data = utilities.explode(self.log, meta_data,
                        expand_header + " Dimension " + str(i+1), reduced_coords_per_dim[i])

            # for other dimensions, just add column
            else:
                expanded_data.add_col(np.concatenate(reduced_coords_per_dim[i]).tolist(),
                    expand_header + " Dimension " + str(i+1))

        # write out .csv file
        expanded_data.to_csv(csv_out, index=csv_no_index, 
                              index_label=csv_index_header,
                              cols=csv_headers)

# if called from the command line display plugin specific command line arguments
if __name__ == "__main__":

    parameter_space = Plugin()
    parameter_space.parser.parse_args()


# Copyright (c) 2021 National Technology and Engineering Solutions of Sandia, LLC.  
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This is a minimal generic plugin that provides no addditional 
# functionality beyond the standard slypi ensemble behavior.

# S. Martin
# 12/16/2020


# standard library imports
import argparse

# 3rd party imports

# local imports
import slypi


# functions for specific operations
# class must be named Plugin
class Plugin(slypi.ensemble.PluginTemplate):

    # initialize command line argument parser
    def __init__(self):

        # describe plugin
        description = "The generic plugin is a minimal plugin that provides " + \
            "no additional support beyond the standard slypi ensemble behavior"

        # set up parser
        super().__init__(description=description)


# if called from the command line display plugin specific command line arguments
if __name__ == "__main__":

    generic = Plugin()
    generic.parser.parse_args()


# Copyright (c) 2013, 2018 National Technology and Engineering Solutions of Sandia, LLC. 
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering 
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this software.

# This script creates a Python distribution wheel for the Slycat python interface, 
# now called slypi, formerly called slycat-web-client.  When run, it copies the latest 
# version of the slycat/web/client files (from the slycat repo) into slypi directory 
# (see note below) before it constructs the pip installable module.
#
# NOTE: 
# 1. You need to have the slycat directory (from github) available as ../slycat
# in order to run the setup.py file (see additional note below).
# 
# S. Martin
# 11/9/2023

# To publish to PyPi, perform the following steps:
#
# $ rm -rf dist
# $ python -m build --sdist --wheel
# $ twine upload dist/*
#
# To publish to testpypi, use:
#
# $ twine upload --repository-url https://test.pypi.org/legacy/ dist/*
#
# The first step builds the distribution, and the second step
# uploads to PyPi.  To install the package from another computer use:
#
# $ pip install slpyi
#
# To install from testpypi, use:
#
# $ pip install --extra-index-url https://testpypi.python.org/pypi slpyi
#   --upgrade --trusted-host testpypi.python.org --proxy wwwproxy.sandia.gov:80
#
# To install locally, use (from this directory):
#
# $ pip install -e .

from shutil import copyfile

# copy slycat.darray and slycat csv parser code into slypi directory. This
# makes slypi a Python package without other Slycat dependencies.
copyfile('../slycat/packages/slycat/darray.py', 'src/slypi/darray.py')
copyfile('../slycat/web-server/plugins/slycat-video-swarm/vs-parse-files.py', 'src/slypi/vs/vs_parse_files.py')

from setuptools import setup

if __name__ == "__main__":
    setup()

# import slypi
# VERSION = slypi.__version__

# # get README.md
# import pathlib

# # directory containing this file
# HERE = pathlib.Path(__file__).parent

# # text of the web-client-readme.txt file
# README = (HERE / "README.md").read_text()

# # create distribution
# import setuptools

# # create Python distribution wheel
# from setuptools import setup

# setup(
#     name="slypi",
#     version=VERSION,
#     description="Slycat python interface utilties for interacting with the Slycat " +
#                 "data analysis and visualization server.",
#     long_description=README,
#     long_description_content_type="text/markdown",
#     url="https://github.com/sandialabs/slypi",
#     author="Shawn Martin",
#     author_email="smartin@sandia.gov",
#     license="Sandia",
#     classifiers=[
#         "Programming Language :: Python :: 3",
#         "Programming Language :: Python :: 3.11",
#     ],
#     packages=setuptools.find_packages(),
#     include_package_data=True,
#     install_requires=["requests", "requests-kerberos",
#                       "numpy", "scikit-learn", 
#                       "pandas", "meshio", "imageio[ffmpeg]",
#                       "matplotlib", "ipyparallel",
#                       "umap-learn", "pyarrow"],
#     entry_points={
#         "console_scripts": [
#             "ps_upload_csv=slypi.ps.upload_csv:main",
#             "dac_upload_gen=slycat.dac.upload_gen:main"
#         ]
#     },
# )
